/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "LoadPara.h"

void PRTerr(const char *); // Show errors
extern TEST TST[TESTnum];
extern int errC; // Errors counter

extern char *filename,*OutFile,*TESTorder;

extern int N,auto_n,OverallTestNumber,IdlePriority,
	NT, // Number of sequences to test
	ShowPvals, // If =1 shows the p-values for each test
	sortKS,sortAD,sortSL, // Sort p-values by KS, AD, SL
	TestName_pv, // If !=0 save test_name[TAB]p-value, else p-value[TAB]test name
	pvalDGT; // P-value digits

extern ULONG DFTmax,DFToptimize;
extern int
	DFTdmin,DFTdmax,DFTdstp, // DFT bit distance
	DFTm0,DFTm1,DFTmmin,DFTmstp, // DFT mode

	AUTWdmin,AUTWdmax,AUTWdstp, // Windowed autocorrelation (distance)
	AUTWmmin,AUTWmmax,AUTWmstp, // Windowed autocorrelation (block size)

	SBLKsmin,SBLKsmax,SBLKsstp, // Short block (block size [bits])
	SBLKdmin,SBLKdmax,SBLKdstp, // Short block (bit distance [bits])

	PERMsmin,PERMsmax,PERMsstp, // Block size [bits]
	PERMtmin,PERMtmax,PERMtstp, // Group size [s-bit numbers]

	RUNsmin,RUNsmax,RUNsstp, // Runs (block size [bit]) 2 <= s <= 32
	RUNrmin,RUNrmax,RUNrstp, // Runs (run length [s-bit numbers]) 2 <= r <= 10

	COUmin,COUmax,COUstp, // Coupon collector
	LBLKmin,LBLKmax,LBLKstp, // Long block
	MAUmin,MAUmax,MAUstp,MAU_n, // Maurer
	NOTMmin,NOTMmax,NOTMstp,NOTMtemp, // NON overlapping template matchings
	SERmin,SERmax,SERstp, // Serial
	RNKmin,RNKmax,RNKstp, // Rank 6x8
	AMLSmin,AMLSmax,AMLSstp, // AMLS
	COLLsmin,COLLsmax,COLLsstp; // Collision (block size) [bits]

extern double KSmin,ADmin,SLmax; // Minimum values for the partial report

static char *section; // Last parameter section scanned
static FILE *par;

static char *PARAtoINT(char *s,int *p)
{
	while(*s && (*s<'0' || *s>'9') && *s!='-') s++; // Discard the leading spaces or tabs
	if(*s==0) {  // Error condition
		sprintf(buf,ERRO"Parameter missing in section: %s\n",section);
		PRTerr(buf); errC++;
	}
	else *p=atoi(s);

	while(*s>='0' && *s<='9' || *s=='-') s++; // Discard the remaining digits of the number (already converted)
	return s; // Update s
}

static void PARA(int *p1,int *p2,int *p3,int *p4)
{
	do fgets(buf,999,par); while(*buf=='#' || !*buf);
	LCASE(buf); char *s=buf,*p=strchr(s,'#'); if(p) *p=0; // Discard the remark
	if(p1) s=PARAtoINT(s,p1);
	if(p2) s=PARAtoINT(s,p2);
	if(p3) s=PARAtoINT(s,p3);
	if(p4)   PARAtoINT(s,p4);
}

static ULONG UM(char *s)
{
	double n=atof(s); ULONG m=1; char *p=strchr(s,' ');
	if(p) {
		while(*++p==' ');
		switch(*p) {
			case 'b': m=1; break; // Bits
			case 'B': m=8; break; // Bytes

			case 'K': m=1024; break;
			case 'M': m=1024*1024; break;
			case 'G': m=1024*1024*1024; break;

			case 'k': m=1000; break;
			case 'm': m=1000*1000; break;
			case 'g': m=1000*1000*1000; break;
		}
		if(p[1]=='B') m*=8; // Bytes
	}
	return ULONG(n*m);
}

int LoadPara(const char *parafile)
{
	par=fopen(parafile,"rb"); int i,j; section=NULL;

	if(!par) {
		sprintf(buf,ERRO"Parameter file \"%s\" not opened.\n",parafile);
		PRTerr(buf); return 1;
	}

	while(fgets(buf,999,par)) {

		char *p=strchr(buf,'#'); if(p) *p=0; // Remark
		if(*buf==0) continue; // Remark

		delete[] section; section=new char[strlen(buf)+1];
		strcpy(section,buf); LCASE(buf);

		// GENERAL
		if(0==strncmp(buf,"[general]",9)) {
	GENERAL:
			if(NULL==fgets(buf,999,par)) {
				if(feof(par)) PRTerr(ERRO"Reached the end of the configuration file while reading the 'General' section.");
				else PRTerr(ERRO"Error while reading the 'General' section.");
				delete[] section; fclose(par); return 1;
			}
			p=strchr(buf,'#'); if(p) *p=0; // Remark

			p=buf; // Find the beginning of the parameter
			while(*p && *p!='[' && (*p<'a' || *p>'z') && (*p<'A' || *p>'Z')) p++;

			if(strlen(p)==0) goto GENERAL;

			if(*p!='[') { // If not next section, assign the parameter

				char *bb=new char[strlen(p)+1]; strcpy(bb,p); LCASE(p);

				if(0==strncmp(p,"n ",2)) n=UM(bb+2); // Sequence length
				else if(0==strncmp(p,"nt",2)) NT=atoi(p+2);
				else if(0==strncmp(p,"auto_n",6)) auto_n=atoi(p+6);
				else if(0==strncmp(p,"showpvals",9)) ShowPvals=atoi(p+9);
				else if(0==strncmp(p,"overalltestnumber",17)) OverallTestNumber=atoi(p+17);
				else if(0==strncmp(p,"testname_pv",11)) TestName_pv=atoi(p+11);
				else if(0==strncmp(p,"pval_digits",11)) pvalDGT=atoi(p+11);
				else if(0==strncmp(p,"idlepriority",12)) IdlePriority=atoi(p+12);

				else if(0==strncmp(p,"sortks",6)) sortKS=atoi(p+6);
				else if(0==strncmp(p,"sortad",6)) sortAD=atoi(p+6);
				else if(0==strncmp(p,"sortsl",6)) sortSL=atoi(p+6);

				else if(0==strncmp(p,"ksmin",5)) KSmin=atof(p+5);
				else if(0==strncmp(p,"admin",5)) ADmin=atof(p+5);
				else if(0==strncmp(p,"slmax",5)) SLmax=atof(p+5);

				else if(0==strncmp(p,"outfile",7)) {
					char *ini=strchr(bb,'\"'),*end; if(ini) end=strchr(ini+1,'\"');
					if(ini==0 || end==0) {
						PRTerr(ERRO"The OutFile file name must be always enclosed in quotation marks.");
						delete[] bb; delete[] section; fclose(par); return 1;
					}
					if(end-ini>1) {
						*end=0; delete[] OutFile;
						OutFile=new char[end-ini]; strcpy(OutFile,ini+1);
					}
				}
				else if(0==strncmp(p,"infile",6)) {
					char *ini=strchr(bb,'\"'),*end; if(ini) end=strchr(ini+1,'\"');
					if(ini==0 || end==0) {
						PRTerr(ERRO"The InFile file name must be always enclosed in quotation marks.");
						delete[] bb; delete[] section; fclose(par); return 1;
					}
					if(end-ini>1) {
						delete[] filename; // In the case filename is declared more than once
						*end=0; filename=new char[end-ini];
						strcpy(filename,ini+1);
					}
				}
				else {
					char *e=p+strlen(p)-1; while(*e==10 || *e==13) *e--=0; 
					sprintf(buf,ERRO"Parameter \"%s\" not recognised in the \"General\" section.\n",p);
					PRTerr(buf); delete[] bb; delete[] section; fclose(par); return 1;
				}
				delete[] bb; goto GENERAL;
			}
			else LCASE(buf); // Next section
		} // end GENERAL

		// ENABLE
		if(0==strncmp(buf,"[enable]",8)) {
			for(i=0;i<TESTnum;i++) TESTorder[i]=-1; // Disable all the tests
			i=0;
			while(fgets(buf,999,par)) {
				if(*buf=='[') { LCASE(buf); break; } // Next section
				char *p=strchr(buf,'#'); if(p) *p=0; // Discard the remark
				if(*buf!='0' && *buf!='1') continue;
				LCASE(buf);
				// Try to find the test name
				for(j=0;j<TESTnum;j++)
					if(0==strncmp(buf+2,TST[j].name,strlen(TST[j].name))) {
						TST[j].flg=*buf-'0'; // Enable/disable the test
						TESTorder[i]=j; break;
					}
				if(j==TESTnum) { // Test not found
					char *err=new char[strlen(buf)+1]; strcpy(err,buf);
					sprintf(buf,ERRO"Test not found in line: %s\n\n",err);
					PRTerr(buf); delete[] err; delete[] section; fclose(par); return 1;
				}
				i++;
			}
		} // end ENABLE

		// PARAMETER sections (buf+1 discards the '[' like in '[DFT]')
		if(0==strncmp(buf+1,TST[dft].name,strlen(TST[dft].name))) {
			fgets(buf,999,par); DFTmax=UM(buf);
			fgets(buf,999,par); DFToptimize=UM(buf);
			PARA(&DFTdmin,&DFTdmax,&DFTdstp,0); // DFT rows
			PARA(&DFTm0,&DFTm1,0,0); // DFT mode
		}
		else if(0==strncmp(buf+1,TST[cou].name,strlen(TST[cou].name))) {
			PARA(&COUmin,&COUmax,&COUstp,0);
		}
		else if(0==strncmp(buf+1,TST[mau].name,strlen(TST[mau].name))) {
			PARA(&MAUmin,&MAUmax,&MAUstp,&MAU_n);
		}
		else if(0==strncmp(buf+1,TST[sblk].name,strlen(TST[sblk].name))) {
			PARA(&SBLKsmin,&SBLKsmax,&SBLKsstp,0); // Block size [bits]
			PARA(&SBLKdmin,&SBLKdmax,&SBLKdstp,0); // Bits distance [bits]
		}
		else if(0==strncmp(buf+1,TST[runs].name,strlen(TST[runs].name))) {
			PARA(&RUNsmin,&RUNsmax,&RUNsstp,0); // Block size [bits]
			PARA(&RUNrmin,&RUNrmax,&RUNrstp,0); // Run length [s-bit numbers]
		}
		else if(0==strncmp(buf+1,TST[perm].name,strlen(TST[perm].name))) {
			PARA(&PERMsmin,&PERMsmax,&PERMsstp,0); // Block size [bits]
			PARA(&PERMtmin,&PERMtmax,&PERMtstp,0); // Group size [bits]
		}
		else if(0==strncmp(buf+1,TST[lblk].name,strlen(TST[lblk].name))) {
			PARA(&LBLKmin,&LBLKmax,&LBLKstp,0);
		}
		else if(0==strncmp(buf+1,TST[autow].name,strlen(TST[autow].name))) {
			PARA(&AUTWdmin,&AUTWdmax,&AUTWdstp,0); // Distance [blocks]
			PARA(&AUTWmmin,&AUTWmmax,&AUTWmstp,0); // Block size [bits]
		}
		else if(0==strncmp(buf+1,TST[notm].name,strlen(TST[notm].name))) {
			PARA(&NOTMmin,&NOTMmax,&NOTMstp,&NOTMtemp);
		}
		else if(0==strncmp(buf+1,TST[srl].name,strlen(TST[srl].name))) {
			PARA(&SERmin,&SERmax,&SERstp,0);
		}
		else if(0==strncmp(buf+1,TST[r68].name,strlen(TST[r68].name))) {
			PARA(&RNKmin,&RNKmax,&RNKstp,0);
		}
		else if(0==strncmp(buf+1,TST[amls].name,strlen(TST[amls].name))) {
			PARA(&AMLSmin,&AMLSmax,&AMLSstp,0);
		}
		else if(0==strncmp(buf+1,TST[coll].name,strlen(TST[coll].name))) {
			PARA(&COLLsmin,&COLLsmax,&COLLsstp,0);
		}
		else if(*buf=='[') {
			sprintf(buf+500,"Test not found: %s",buf); PRTerr(buf+500);
		}
		// end PARAMETER section

	} // while(fgets(buf,999,par))

	delete[] section; fclose(par); return 0;
}
