/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

// DONT'USE THE INTEL COMPILER 8.0 (the executable will be much slower)

#define VERSION "RaBiGeTe v050711 - by Cristiano\n"

	#ifdef __GNUC__
#undef USE_PERF_COUNTER
#include <glob.h>
	#else
#define USE_PERF_COUNTER // Performance counter instead of clock()
#define WIN32_LEAN_AND_MEAN
#define NOSOUND
#include <windows.h> // Performance counter and SetPriorityClass
#include <io.h> // _findfirst
#include <conio.h> // _kbhit

	#endif

#include "prova.h"

#include "defi.h"
#include "LoadPara.h"
#include <malloc.h>
#include <time.h>

// TESTS
void DFT(void);
void MAURER(void);
void SHORTBLK(void);
void LONGBLK(void);
void TWOBIT(void);
void BLKGAP(void);
void RUNS(void);
void CUSUM(void);
void SOB(void);
void NOTM(void);
void RANK6_8  (void);
void RANK31_31(void);
void RANK32_32(void);
void SERIAL(void);
void GCD(void);
void BDAY(void);
void AUTOw(void);
void PERM(void);
void COUPON(void);
void AMLS(void);
void COLLISION(void);

// Deallocate the resources used by some test
void DFTfree(void);
void RUNSfree(void);
void COLLISIONfree(void);

TEST TST[] = { // Don't change neither the order nor the names (the order must match the enum in the module "LoadPara.h")
	1,DFT,         "dft",
	1,MAURER,      "maurer",
	1,SHORTBLK,    "short blocks",
	1,LONGBLK,     "long blocks",
	1,TWOBIT,      "serial two-bit",
	1,BLKGAP,      "blocks'n'gaps",
	1,RUNS,        "runs",
	1,CUSUM,       "cumulative sum",
	1,SERIAL,      "serial",
	1,NOTM,        "non overlapping template matchings",
	1,RANK6_8,     "rank 6x8",
	1,RANK31_31,   "rank 31x31",
	1,RANK32_32,   "rank 32x32",
	1,SOB,         "sparse occupancy bitstream",
	1,GCD,         "gcd",
	1,BDAY,        "tough birthday spacings",
	1,AUTOw,       "windowed autocorrelation",
	1,PERM,        "permutation",
	1,COUPON,      "coupon collector",
	1,AMLS,        "amls",
	1,COLLISION,   "collision"
};

//	#ifdef _PROVA_
//#define DEL delete[] buf; delete[] filename; delete[] OutFile; delete[] TESTorder; delete[] LOGf; if(ResFile) fclose(ResFile); system("pause");
//	#else
#define DEL delete[] buf; delete[] filename; delete[] OutFile; delete[] TESTorder; delete[] LOGf; if(ResFile) fclose(ResFile);
//	#endif
	

#define ERR(num,msg) { printf("%s: %s\n\n",TST[num].name,msg); errC++; } // Error message

void PRTerr(const char *e) // Show errors in LoadPara
{
	puts(e);
}

// -- LoadPara ----------------------

int LoadPara(const char *parafile);
int errC; // Errors counter

void LCASE(char *s)
{
	while(*s) { if(*s>='A' && *s<='Z') *s+=32; s++; }
}

char *filename=NULL,*OutFile=NULL,*TESTorder;

// Test parameters
ULONG DFTmax,DFToptimize;
int
	DFTdmin,DFTdmax,DFTdstp, // DFT columns
	DFTm0,DFTm1,DFTmmin,DFTmstp, // DFT mode

	AUTWdmin,AUTWdmax,AUTWdstp, // Windowed autocorrelation (distance)
	AUTWmmin,AUTWmmax,AUTWmstp, // Windowed autocorrelation (block size)

	SBLKsmin,SBLKsmax,SBLKsstp, // Short block (block size [bits])
	SBLKdmin,SBLKdmax,SBLKdstp, // Short block (bit distance [bits])

	PERMsmin,PERMsmax,PERMsstp, // Permutation (block size [bits])
	PERMtmin,PERMtmax,PERMtstp, // Permutation (group size [s-bit numbers])

	RUNsmin,RUNsmax,RUNsstp, // Runs (block size [bit]) 2 <= s <= 32
	RUNrmin,RUNrmax,RUNrstp, // Runs (run length [s-bit numbers]) 2 <= r <= 10

	COUmin,COUmax,COUstp, // Coupon collector
	LBLKmin,LBLKmax,LBLKstp, // Long block (block size) 8<= m < n
	MAUmin,MAUmax,MAUstp,MAU_n, // Maurer
	NOTMmin,NOTMmax,NOTMstp,NOTMtemp, // NON overlapping template matchings
	SERmin,SERmax,SERstp, // Serial
	RNKmin,RNKmax,RNKstp, // Rank 6x8
	AMLSmin,AMLSmax,AMLSstp, // AMLS block length [kbits]
	COLLsmin,COLLsmax,COLLsstp, // Collision (block size)

	N,auto_n,OverallTestNumber,IdlePriority,
	NT, // Number of sequences to test
	ShowPvals, // If =1 shows the p-values for each test
	sortKS,sortAD,sortSL, // Sort p-values by KS, AD, SL
	TestName_pv, // If !=0 save test_name[TAB]p-value, else p-value[TAB]test name
	pvalDGT; // P-value digits

ULONG n=0; // Sequence length

double KSmin,ADmin,SLmax; // Minimum values for the partial report

// -- end LoadPara ------------------

ULONG *bits=NULL; // Sequence of bits
char *buf=NULL; // General purpose

double *LOGf=NULL; // Logarithm table for the Windowed Autocorrelation test
// C(n,i) / 2^n
double _FASTCALL_ Cprob(int n,int i) { return exp(LOGf[n]-LOGf[i]-LOGf[n-i]-n*M_LN2); }

void RBG(void);

int pvN=0; // Number of p-values in pv
double *pv=NULL; // P-values obtained from the tests

static int BigFailure; // If =1 big failure in one or more tests

static FILE *filebits=NULL, // Input file
	*ResFile=NULL; // Output (results) file. OutFile (_RaBiGeTe.txt)

static int eleN; // Number of bits file loaded

void PRT(const char *s)
{
	printf(s); fputs(s,ResFile);
}

void _FASTCALL_ SAVE(char *s,double p)
{
	static int pvA=0;
	if(pvN==pvA) {
		pvA+=100; double *t=new double[pvA];
		for(int i=0;i<pvN;i++) t[i]=pv[i];
		delete[] pv; pv=t;
	}

	char tmp[256]; pv[pvN++]=p;
	if(TestName_pv) sprintf(tmp,"%s\t%.*f\n",s,pvalDGT,p); // test_name[TAB]p-value
	else sprintf(tmp,"%.*f\t%s\n",pvalDGT,p,s); // p-value[TAB]test_name
	fputs(tmp,ResFile);
	if(ShowPvals) printf(tmp);
	else if(p==-1 && BigFailure==0) { // Big failure
		BigFailure=1; puts("  Warning: big failure for one or more tests.\n  Please, see the output file for a detailed report.\n");
	}
}

static char *TIMEtoHMS(char *p,double x,int dgt)
{
	double M=1; if(dgt) M=pow(10.,dgt);
	x=int(x*M+.5)/M/3600.;
	int d, m; d = (int)x; x -= d; x *= 60; m = (int)x; x -= m; x *= 60;
	if(dgt) sprintf(p,"%d:%02d:%04.*f",d,m,dgt,x);
	else {
		if(d) sprintf(p,"%2dh %2dm",d,m);
		else if(m) sprintf(p,"%2dm %2ds",m,int(x));
		else sprintf(p,"%2ds",int(x));
	}
	return p;
}

// Sort the p-values "pv" and the index "id"
void _FASTCALL_ SORT(double *pv,ULONG *id,int s,int d,const int flg)
{
	int i=s,j=d; double p=pv[(s+d)/2];
	do {
		if(flg) { // Ascending order
			while(pv[i]<p && i<d) i++;
			while(p<pv[j] && j>s) j--;
		} else { // Descending order
			while(pv[i]>p && i<d) i++;
			while(p>pv[j] && j>s) j--;
		}
		if(i<=j) {
			double t=pv[i]; pv[i]=pv[j]; pv[j]=t;
			ULONG tu=id[i]; id[i]=id[j]; id[j]=tu;
			i++; j--;
		}
	} while(i<=j);

	if(s<j) SORT(pv,id,s,j,flg);
	if(i<d) SORT(pv,id,i,d,flg);
}


// Sort the files in the selected dir
void _FASTCALL_ SORT(char **ele,int s,int d)
{
	int i=s,j=d; char *p=ele[(s+d)/2];
	do {
#ifdef __GNUC__
		while(strcmp(ele[i],p)<0 && i<d) i++;
		while(strcmp(p,ele[j])<0 && j>s) j--;
#else
		while(stricmp(ele[i],p)<0 && i<d) i++; // Perform a lowercase comparison of strings
		while(stricmp(p,ele[j])<0 && j>s) j--;
#endif
		if(i<=j) {
			char *t=ele[i]; ele[i]=ele[j]; ele[j]=t;
			i++; j--;
		}
	} while(i<=j);

	if(s<j) SORT(ele,s,j);
	if(i<d) SORT(ele,i,d);
}

// Stores in 'ele' the files specified in 'dir' (with wildcards)
// Passing ele=NULL, only C will be calculated (no storing in ele)

#ifdef __GNUC__

int FINDfiles(char *dir,char **ele,ULONG &S)
{
	glob_t gl; unsigned int i;

	if (glob(dir, 0, NULL, &gl) == GLOB_NOMATCH) return 0;
	for (i=0; i<gl.gl_pathc; i++) {
		if (ele) {
			ele[i]=new char[strlen(gl.gl_pathv[i])+1];
			strcpy(ele[i],gl.gl_pathv[i]);
		}
		// S is the sum of the files ele[]. Write the right instruction
		S+= FILE_SIZE_OF ele[i]; // CHANGE THIS!
    }
	globfree(&gl); return i;
}

#else

int FINDfiles(char *dir,char **ele,ULONG &S)
{
    struct _finddata_t fd; intptr_t hFile; int C=0;

    if((hFile=_findfirst(dir,&fd))== -1L) return 0; // File not found

	do { //printf("%d  %s\n",fd.attrib & _A_SUBDIR,fd.name);
		if((fd.attrib & _A_SUBDIR)==0) { // No dirs
			if(ele) {
				ele[C]=new char[strlen(fd.name)+1]; strcpy(ele[C],fd.name);
			}
			C++; S+=fd.size;
		}
	} while(_findnext(hFile,&fd)== 0);

    _findclose(hFile); return C;
}

#endif

//	Fills the vector 'bits' with n bytes loaded from the files 'ele' located
//	in the directory 'dir'.
int FILLbits(char *dir,char **ele,const int eleMAX,UCHAR *bits,const int n)
{
	int letti=0;
		
	if(filebits==NULL) {
LAB:
		strcpy(buf,dir); strcat(buf,ele[eleN]); //printf("<%s>\n",buf);
		if((filebits=fopen(buf,"rb"))==NULL) {
			printf(ERRO"Bits file \"%s\" not opened.\n",buf); return -2;
		}
	}

	int necess=n-letti,
		rb=fread(bits+letti,1,necess,filebits); letti+=rb;

	if(rb!=necess) { //puts("\trb != necess");

		if(feof(filebits)) { // File exhausted
			fclose(filebits); filebits=NULL;
			if(++eleN<eleMAX) goto LAB; // Load another file
		}
		else {
			printf(ERRO"Bits file \"%s\" not loaded.\n",ele[eleN]);
			fclose(filebits);
		}
		return rb; // Error: return the number of bytes loaded
	}

    return -1; // No error
}


int main(int argc, char* argv[])
{
		#ifdef USE_PERF_COUNTER
	LARGE_INTEGER freq,RunTime,t0,tn;
	QueryPerformanceFrequency(&freq); QueryPerformanceCounter(&RunTime);
		#else
	clock_t t0,RunTime=clock();
		#endif

	buf=new char[1000]; // General purpose buffer

	puts(VERSION);

		#ifndef _PROVA_
	if(argc!=2) {
		puts("USAGE:");
		puts("  RaBiGeTe DHcheck:      the program executes a test intended to check the\n"
			 "                         diehard tests.\n"
			 "  RaBiGeTe \"param_file\": the program runs configured with the parameters\n"
			 "                         in param_file.\n");
		delete[] buf; /*system("pause");*/ return 1;
	}
		#endif

	int i,j;
	OutFile=new char[15]; strcpy(OutFile,"_RaBiGeTe.txt"); // Default output file
	TESTorder=new char[TESTnum]; // The user can change the order of execution

	MAUmin=2; MAUmax=16; MAUstp=1; MAU_n=1000; // Maurer
	LBLKmin=1024; LBLKmax=8192; LBLKstp=-2; // Long block block size 8<= m < n
	NOTMmin=2; NOTMmax=16; NOTMstp=-1; NOTMtemp=3; // Non overlapping template matchings
	SERmin=2; SERmax=16; SERstp=-1; // Serial
	RNKmin=0; RNKmax=24; RNKstp=1; // Rank 6x8
	COUmin=2; COUmax=9; COUstp=1; // Coupon collector
	AMLSmin=4; AMLSmax=32768; AMLSstp=-1; // AMLS
	COLLsmin=1; COLLsmax=31; COLLsstp=1; // Collision block size [bits] s <= 31

	PERMsmin=1; PERMsmax=16; PERMsstp=-1; // Block size [bits]
	PERMtmin=2; PERMtmax=12; PERMtstp=1; // Group size [s-bit numbers] 2 <= t <= 12

	SBLKsmin=8; SBLKsmax=20; SBLKsstp=-1; // Short block block size [bits]
	SBLKdmin=1; SBLKdmax=4096; SBLKdstp=-4; // Short block bit distance [bits]

	AUTWmmin=1; AUTWmmax=16; AUTWmstp=-2; // Windowed auto block size [bits]
	AUTWdmin=16; AUTWdmax=4096; AUTWdstp=-2; // Windowed auto distance [blocks]

	RUNsmin=2; RUNsmax=3; RUNsstp=1; // Runs block size [bit]) 2 <= s <= 32
	RUNrmin=2; RUNrmax=3; RUNrstp=1; // Runs run length [s-bit numbers] 2 <= r <= 10

	DFTmax=3*1024*1024; DFToptimize=2*1024*1024;
	DFTdmin=1; DFTdmax=512; DFTdstp=-3; // DFT rows
	DFTm0=1; DFTm1=1; // DFT mode

	TestName_pv=0; // If !=0 save test_name[TAB]p-value, else p-value[TAB]test name
	pvalDGT=6; auto_n=0; ShowPvals=0; OverallTestNumber=0; IdlePriority=1;
	sortKS=1; sortAD=0; sortSL=0;
	KSmin=.0005; ADmin=0; SLmax=100;

	NT=5; // Number of sequences to test
	n=2*1024*1024; // Sequence length [bits]
	if(NT>=10) ShowPvals=0;

	for(i=0;i<TESTnum;i++) { TST[i].flg=0; TESTorder[i]=i; }
	//for(i=0;i<TESTnum;i++) { TST[i].flg=1; TESTorder[i]=i; }

		#if 0 // Test a file
	NT=100; n=0; filename=new char[900];
	strcpy(filename,"C:\\PRG_VC2005\\ApoditticoECC\\release\\_ECrnd_MIO_?.bin");
		#endif

	TST[coll ].flg=	1; // Collision
/*
	TST[dft  ].flg=	1; // DFT
	TST[amls ].flg=	1; // AMLS
	TST[cus  ].flg=	1; // CuSum
	TST[sblk ].flg=	1; // Short blocks
	TST[runs ].flg=	1; // Runs
	TST[perm ].flg=	1; // Permutation
	TST[lblk ].flg=	1; // Long blocks
	TST[blkg ].flg=	1; // Blocks'n'gaps
	TST[twob ].flg=	1; // Two-bit
	TST[notm ].flg=	1; // NOTM
	TST[autow].flg=	1; // AUTOw
	TST[srl  ].flg=	1; // Serial
	TST[mau  ].flg=	1; // Maurer
	TST[cou  ].flg=	1; // Coupon collector
	TST[sob  ].flg=	1; // SOB

	TST[gcd  ].flg=	1; // GCD
	TST[bday ].flg=	1; // Bday
	TST[r68  ].flg=	1; // Rank 6x8
	TST[r31  ].flg=	1; // Rank 31x31
	TST[r32  ].flg=	1; // Rank 32x32
	//*/

	errC=0; // Errors counter

		#ifndef _PROVA_
	if(argc>1) {
		LCASE(argv[1]);
		if(0==strcmp(argv[1],"dhcheck")) { // DH test vector (Set z1=z2=z3=z4=9999 in lfsr113)

			RNKmin=0; RNKmax=24; RNKstp=1; // Rank 6x8
			auto_n=1; ShowPvals=1; NT=1; TestName_pv=0; pvalDGT=8;

			for(i=0;i<TESTnum;i++) { TST[i].flg=0; TESTorder[i]=i; }
			TST[r68].flg=1;	// Rank	6x8
			TST[r31].flg=1;	// Rank	31x31
			TST[r32].flg=1;	// Rank	32x32
			TST[sob].flg=1;	// SOB
			TST[gcd].flg=1;	// GCD
			TST[bday].flg=1; //	Bday
		}
		else if(LoadPara(argv[1])) { DEL; return 1; } // User's prefs
	}
		#endif

#ifndef __GNUC__
	/* Linux users: here we could call getpriority() and setpriority() */
	if(IdlePriority)
		if(!SetPriorityClass(GetCurrentProcess(),IDLE_PRIORITY_CLASS)) puts("\n\tPriority not changed\n");
#endif


	if((ResFile=fopen(OutFile,"w+b"))==NULL) {
		printf(ERRO"Output file \"%s\" not created.\n",OutFile);
		DEL; return 1;
	}

	fputs(VERSION,ResFile);

	char **ele; int eleC=0; ULONG TOTlen; // Total files length

	if(filename) {
		eleC=FINDfiles(filename,NULL,TOTlen); // Counts the files
		if(eleC==0) {
			printf(ERRO"No matching file for:\n%s\n",filename);
			DEL; return 1;
		}
		ele=new char*[eleC]; TOTlen=0; FINDfiles(filename,ele,TOTlen); SORT(ele,0,eleC-1);
		if(eleC==1) sprintf(buf,"File: %s\n",filename);
		else sprintf(buf,"Files: %s (%d files)\n",filename,eleC);
		PRT(buf); char *p=strrchr(filename,'\\');
		if(p) p[1]=0; else *filename=0; // No dir
		if(n==0) n=TOTlen/NT*8; // Set n to test all the available bits (TOTlen is in bytes)
	} else if(n==0) {
		printf(ERRO"Sequence length = 0 is only allowed if an input file is specified.\n");
		DEL; return 1;
	}



	if(pvalDGT<1) {
		puts(ERRO"pval_digits must be > 0.\n"); DEL; return 1; 
	}


	// ---------------------------------------------------------------------------
	// PARAMETER CHECKING (the tests with the invalid parameters will be disabled)
	// ---------------------------------------------------------------------------

	if(TST[amls].flg) {
		if(AMLSmin<2 || AMLSmax>32768 || AMLSmin>AMLSmax || AMLSstp>-1) {
			ERR(amls,"min >= 2, max <= 32768, min <= max and step < 0"); TST[amls].flg=0;
		}
		int blk=AMLSmin; // Check the form 2^x for the block size
		while(blk>2) {
			if(blk&1) {
				ERR(amls,"the block size must be of the form 2^x"); TST[amls].flg=0; break;
			}
			blk>>=1;
		}
	}

	if(TST[autow].flg) // Windowed Auto: set up the logarithm table
		if( AUTWdmin<1 || AUTWdmin>AUTWdmax || AUTWdstp==0 ||
			AUTWmmin<1 || AUTWmmin>AUTWmmax || AUTWmstp==0) {
			ERR(autow,"min > 0, min <= max and step != 0"); TST[autow].flg=0;
		} else { // Find the biggest m using the step
			int m; for(i=AUTWmmin; i<=AUTWmmax; ) { m=i; INCR(i,AUTWmstp); } AUTWmmax=m;
			LOGf=new double[AUTWmmax+1]; LOGf[0]=LOGf[1]=0;
			double f=0; for(int i=2;i<=AUTWmmax;i++) { f+=log((double)i); LOGf[i]=f; }
		}

	if(TST[coll].flg) { // Collision
		if(COLLsmin<1 || COLLsmin>COLLsmax || COLLsstp==0) {
			ERR(coll,"min > 0, min <= max and step != 0"); TST[coll].flg=0;
		}
		FILE *lim=fopen("DAT_collision_limits.txt","rb");
		if(lim) { // Check NT max
			int nt;
			while(fgets(buf,999,lim)) {
				nt=atoi(buf); // NT max for the following set of bls limits
				while(fgets(buf,999,lim)) if('#'==*buf) break; // Set end
				if(NT<=nt) break; // Found a good set of bls limits for the current NT
			}
			fclose(lim);

			if(NT>nt) { // NT is too big; you need to calculate the limits for the required NT
				printf("The maximum number of sequences currently supported by the\n"
					"COLLISION test is %d.\nYou chosen %d sequences.\n\n",nt,NT);
				TST[coll].flg=0;
			}
		} else {
			printf(ERRO"%s: file 'DAT_collision_limits.txt' not found.\n\n",TST[coll].name);
			TST[coll].flg=0; errC++;
		}
	}

	if(TST[runs].flg) {
		if(RUNsmin<2 || RUNsmax>32 || RUNsmin>RUNsmax || RUNsstp==0
		|| RUNrmin<2 || RUNrmax>10 || RUNrmin>RUNrmax || RUNrstp==0) {
			ERR(runs,"block size min >= 2, max <= 32, min <= max and step != 0");
			ERR(runs,"runs min >= 2, max <= 10, min <= max and step != 0"); TST[runs].flg=0;
		}
	}

	if(TST[cou].flg) {
		if(COUmin<2 || COUmin>COUmax || COUstp==0) {
			ERR(cou,"min > 1, min <= max and step != 0"); TST[cou].flg=0;
		}
		else {
			FILE *c=fopen("DAT_coupon_prob.txt","rb");
			if(c) fclose(c);
			else {
				printf(ERRO"%s: file 'DAT_coupon_prob.txt' not found.\n\n",TST[cou].name);
				TST[cou].flg=0; errC++;
			}
		}
	}

	if(TST[dft].flg) {
		if(DFTdmin<1 || DFTdmin>DFTdmax || DFTdstp==0) {
			ERR(dft,"min > 0, min <= max and step != 0"); TST[dft].flg=0;
		}
		if(DFTm0==0 && DFTm1==0) {
			ERR(dft,"At least 1 mode must be enabled."); TST[dft].flg=0;
		}
		// Set mode min and step
		DFTmmin=1; DFTmstp=2; if(DFTm0) { DFTmmin=0; if(DFTm1) DFTmstp=1; }
	}

	if(TST[mau].flg && (MAUmin<2 || MAUmax>16 || MAUmin>MAUmax || MAUstp==0 || MAU_n<40)) {
		ERR(mau,"min > 1, max<17, min <= max, step != 0 and N >= 40"); TST[mau].flg=0;
	}

	if(TST[sblk].flg) {
		if(SBLKsmin<1 || SBLKsmax>31 || SBLKsmin>SBLKsmax || SBLKsstp==0
		|| SBLKdmin<1 || SBLKdmin>SBLKdmax || SBLKdstp==0) {
			ERR(sblk,"block size min > 0, max < 32, min <= max and step != 0");
			printf("%s: distance min > 0, min <= max and step != 0\n\n",TST[sblk].name);
			TST[sblk].flg=0;
		}
		// Short blk 1_1 = Serial 1_1 = Serial 1_2
		if(SBLKsmin==1 && SBLKdmin==1 && SERmin==1) SERmin=2;
	}

	if(TST[srl].flg && (SERmin<1 || SERmin>SERmax || SERstp==0)) {
		ERR(srl,"min > 0, min <= max and step != 0"); TST[srl].flg=0;
	}

	// PERMtmax limited to 12. 13!=6,227,020,800
	if(TST[perm].flg && (PERMsmin<1 || PERMsmax>32 || PERMsmin>PERMsmax || PERMsstp==0
		|| PERMtmin<2 || PERMtmax>12 || PERMtmin>PERMtmax || PERMtstp==0)) {
		ERR(perm,"block size min > 0, max <= 32, min <= max and step != 0");
		ERR(perm,"group size min > 1, max <= 12, min <= max and step != 0");
		TST[perm].flg=0;
	}

	if(TST[lblk].flg && (LBLKmin<8 || LBLKmin>LBLKmax || LBLKstp==0)) {
		ERR(lblk,"min >= 8, min <= max and step != 0"); TST[lblk].flg=0;
	}

	if(TST[notm].flg && (NOTMmin<2 || NOTMmax>31 || NOTMmin>NOTMmax || NOTMstp==0)) {
		ERR(notm,"min > 1, max < 32, min <= max and step != 0"); TST[notm].flg=0;
	}

	if(TST[r68].flg && (RNKmin<0 || RNKmax>24 || RNKmin>RNKmax || RNKstp<1)) {
		ERR(r68,"min >= 0, max <= 24, min <= max and step > 0"); TST[r68].flg=0;
	}


	if(errC) {
		puts("\n  ERROR in the test parameters.\nThe tests with the invalid parameters will be disabled.\n");
		system("pause"); errC=0;
	}



	if(auto_n) { // Change n to run the enabled tests
		ULONG nb=n;
		if(TST[dft  ].flg) n=max(1000ul*DFTdmax,n); // DFT
		if(TST[mau  ].flg) n=max(MAU_n*(1ul<<MAUmax)*MAUmax,n); // Maurer
		if(TST[sblk ].flg) n=max(ULONG(SBLKsmax)*SBLKdmax, max(5ul*(1ul<<SBLKsmax)*SBLKsmax,n)); // Short blocks
		if(TST[notm ].flg) n=max( (5*(1ul<< NOTMmax)+NOTMmax-1)*3 ,n); // NOTM
		if(TST[blkg ].flg) n=max(n,2000); // BlkGap
		if(TST[r68  ].flg) n=max(600000ul*32,n); // Rank 6x8
		if(TST[r31  ].flg) n=max(1240000ul*32,n); // Rank 31x31
		if(TST[r32  ].flg) n=max(1280000ul*32,n); // Rank 32x32
		if(TST[sob  ].flg) n=max(67108889ul,n); // SOB
		if(TST[gcd  ].flg) n=max(20000000ul*32,n); // GCD
		if(TST[bday ].flg) n=max(2048000ul*32,n); // Bday
		if(TST[autow].flg) { // Windowed Auto
			n=max(n,ULONG(AUTWdmax)*AUTWmmax);
			n=max(n,AUTWmmax*ULONG(ceil(5./Cprob(AUTWmmax,AUTWmmax/2)+AUTWdmax+1ul)));
		}
		if(TST[runs].flg) { // Runs
			for(int s=RUNsmin; s<=RUNsmax; INCR(s,RUNsstp)) {
				double N=1ul<<s; if(s==32) N=4294967296.;
				for(int r=RUNrmin; r<=RUNrmax; INCR(r,RUNrstp)) {
					double p=exp(lgam(N+1.)-lgam(r+1.)-lgam(N-r+1.)-r*log(N));
					n=max(n,ULONG(5./.27*s/p)+1ul);
				}
			}
		}
		if(TST[coll].flg) { // Collision
			FILE *lim=fopen("DAT_collision_limits.txt","rb");
			int s,nt,smin,smax; ULONG B[32];
			while(fgets(buf,999,lim)) { // The 1st line of each set is NTmax
				smin=99; smax=0; nt=atoi(buf); // NT max for this set of blkS
				memset(B,0,sizeof(B)); // bls only in the allowed block size range
				while(fgets(buf,999,lim)) {
					if('#'==*buf) break; // Set end
					s=atoi(buf); // Block size
					if(s<smin) smin=s; // Smallest allowed blkS for the current NT
					if(s>smax) smax=s; // Biggest allowed blkS for the current NT
					B[s]=atof(strchr(buf,'\t')+1); // bls min for the given block size
				}
				if(NT<=nt) { // The set just loaded is good for the current NT
					COLLsmin=max(COLLsmin,smin); // blkS in the allowed range
					COLLsmax=min(COLLsmax,smax); // blkS in the allowed range
					ULONG bls=0,blkS; // Find bls using the step
					for(s=COLLsmin; s<=COLLsmax; INCR(s,COLLsstp))
						if(B[s]>bls) { bls=B[s]; blkS=s; } // Store the bls for the given blkS
					n=max(n, bls*blkS); break;
				}
			}
			fclose(lim);
		}

		if(nb!=n) printf("\n\tSequence length increased to %u bits\n\n",n);

	} // if(auto_n)


	// ------------------------------------------------
	// CHECK for n (disable the test if n < n_required)
	// ------------------------------------------------

	if(TST[autow].flg) {
		ULONG AWnmin=max(AUTWdmin*AUTWmmin, AUTWmmin*ULONG(ceil(5./Cprob(AUTWmmin,AUTWmmin/2)+AUTWdmin+1ul)));
		if(n<AWnmin) {
			printf("%s: n >= %d\n\n",TST[autow].name,AWnmin); TST[autow].flg=0; errC++;
		}
	}

	if(TST[sblk].flg && SBLKdmax>n/SBLKsmin) {
		printf("%s: distance min <= %d\n\n",TST[sblk].name,n/SBLKsmin);
		TST[sblk].flg=0; errC++;
	}

	if(TST[srl].flg && SERmin>log((double)n)/M_LN2-2) {
		printf("%s: min <= %d\n\n",TST[srl].name,int(log((double)n)/M_LN2-2));
		TST[srl].flg=0; errC++;
	}

	if(TST[lblk].flg && LBLKmax>n) { ERR(lblk,"max <= n"); TST[lblk].flg=0; }

	if(errC) system("pause");


	int nt=0; for(i=0;i<TESTnum;i++) nt+=TST[i].flg;
	if(nt==0) { 
		puts(ERRO"No test enabled.\n"); DEL; return 1; 
	}


    // Header (n, date and time)
	int len=sprintf(buf,"n= %u bits",n); char suff=' ';
	double nn=n; if(nn>1023) { nn/=1024; suff='K'; }
	if(nn>1023) { nn/=1024; suff='M'; }
	if(nn>1023) { nn/=1024; suff='G'; }
	if(nn!=n) len+=sprintf(buf+len," (%.3f %cb)     ",nn,suff);
	else len+=sprintf(buf+len,"     ");

	time_t ltime; time(&ltime); struct tm *lt=localtime(&ltime);
	strftime(buf+len,99,"Started: %Y-%m-%d %H:%M:%S\n\n",lt); PRT(buf);

	bits=new ULONG[n/32+1]; int NR=0; // Rows
	const int NTorig=NT; // NT set by the user

	fpos_t INItst; fgetpos(ResFile,&INItst); // Position of the first test

	// Do the tests
	for(N=1;N<=NT;N++) {

		if(filename) {
			int rb=FILLbits(filename,ele,eleC,(UCHAR*)bits,n/8);
			if(rb!=-1) {
				if(rb!=-2) {
					printf("\n\n\t!! LOADING ERROR !!\nFile \"%s%s\"\n",filename,ele[eleN-1]);
					printf("Only %u bits have been read instead of %u\n\n",rb*8,n);
					puts("\n\tPress enter to continue"); getchar();
				}
				NT=N-1; break; // Terminate the test
			}
		}
		else RBG(); // Fill the buffer 'bits' with n bits

		sprintf(buf,"Sequence #%d\n",N); fputs(buf,ResFile);
		if(ShowPvals) printf(buf); else if(N==1) printf("Testing the sequence #1...\r");

		for(i=0;i<TESTnum;i++) // Run the enabled tests in the selected order
			if(TESTorder[i]!=-1) if(TST[TESTorder[i]].flg) TST[TESTorder[i]].func();

		if(N==1) NR=pvN;

		// PARTIAL P-VALUES
		if((KSmin>0 || ADmin>0 || SLmax<100) && N>2) {
			int pvC; double *pvt=new double[pvN]; // Temporary p-values
			for(i=0;i<NR;i++) {
				double pvKS,pvAD,pvSL;
				pvC=0; for(int j=0;j<N;j++) if(pv[i+j*NR]!=-1) pvt[pvC++]=pv[i+j*NR];
				if(pvC>1) {
					sort_pv(pvt,pvC);
					pvKS=KStest(pvt,pvC); pvAD=ADtest(pvt,pvC); pvSL=SLtest(pvt,pvC);
				}
				else { pvKS=pvAD=-1; pvSL=100; }
				if(pvKS<KSmin || pvAD<ADmin || pvSL>SLmax)
					printf("Low overall p-value for sequence #%d: KS= %.4f  AD= %.4f  SL=%5.1f\n",N,pvKS,pvAD,pvSL);
			}
			delete[] pvt;
		}

		if(NT>2) { // Only if there are at least 3 sequences to test

			if(N==1) { // The tests have been initialized; this is T0
					#ifdef USE_PERF_COUNTER
				QueryPerformanceCounter(&t0);
					#else
				t0=clock();
					#endif
			} else {

					#ifdef USE_PERF_COUNTER // Elapsed time
				QueryPerformanceCounter(&tn); double T=(tn.QuadPart-t0.QuadPart)/double(freq.QuadPart);
					#else
				double T=(clock()-t0)/(double)CLOCKS_PER_SEC;
					#endif

				if(!ShowPvals) printf(" #%d",N); // Sequence #
				double TR=T/(N-1)*(NT-N);
				printf(" Elapsed: %s    Left: %s",TIMEtoHMS(buf,T,0),TIMEtoHMS(buf+20,TR,0));
				time(&ltime); ltime+=(time_t)TR; lt=localtime(&ltime);
				strftime(buf,99,"    Date end: %Y-%m-%d %H:%M    \r",lt); printf(buf);
			}
		}

		fputs("\n",ResFile); if(ShowPvals) puts("\n");

			#ifndef __GNUC__
		if(_kbhit()) if(_getch()==27) NT=N; // Escape pressed
			#endif

	} // for(N=1;N<=NT;N++) TEST END

	if(TST[dft].flg) DFTfree(); // Free the memory used by the DFT test
	if(TST[runs].flg) RUNSfree(); // Free the memory used by the RUNS test
	if(TST[coll].flg) COLLISIONfree(); // Free the memory used by the COLLISION test

	if(eleC) { for(i=0;i<eleC;i++) delete[] ele[i]; delete[] ele; }

	printf("%80s\r",""); // Delete "Elapsed... Left..."


	// RESULTS
	// -------

	int pvC; // Valid p-values

	double *pvt=new double[pvN]; // Temporary p-values
	for(i=0;i<pvN;i++) pvt[i]=pv[i];


	if(NT>2) { // Overall values only if there are at least two tests (NT>1)

		char **RowName=new char*[NR]; int maxlen=0;
		fpos_t pos; fgetpos(ResFile,&pos); // Current file position
		fsetpos(ResFile,&INItst); // Beginning of the p-values 
		fgets(buf,999,ResFile); // Discard the header and "Test#1"

		for(i=0;i<NR;i++) { // Read the name of each test
			fgets(buf,999,ResFile);
			char *p=strchr(buf,'\t'); // Find tab separator

			if(TestName_pv) { // test_name[TAB]p-value
				RowName[i]=new char[p-buf+1]; // Allocate space for the name
				strncpy(RowName[i],buf,p-buf); RowName[i][p-buf]=0;
				maxlen=max(maxlen,p-buf);
			}
			else { // p-value[TAB]test_name
				RowName[i]=new char[strlen(p)]; // Allocate space for the name
				p[strlen(p)-1]=0; strcpy(RowName[i],p+1); // Save the name w/o '\n'
				maxlen=max(maxlen,strlen(p)-1);
			}
		}
		fsetpos(ResFile,&pos); // Restore the position


		if(BigFailure) { // Show the tests with big failure
			int c=0; PRT("Tests with big failure:\n");
			for(i=0;i<NR;i++)
				for(int j=0;j<NT;j++)
					if(pv[i+j*NR]==-1) {
						sprintf(buf,"%-25s",RowName[i]); PRT(buf); if(++c%3==0) PRT("\n");
						break;
					}
			PRT("\n"); if(c%3) PRT("\n");
		}


		// Used for the overall p-values
		double *pvKS=new double[max(NT,NR)],
			*pvAD=new double[max(NT,NR)],
			*pvSL=new double[max(NT,NR)],SLsum;

		int OvTST; // Counter of valid tests

		// Overall Sequence# p-value (this p-value is useless)
		// ---------------------------------------------------

		if(NR>1 && OverallTestNumber==1) {

			sprintf(buf,"SEQUENCE#\tKS\tAD\tSL\n"); PRT(buf); SLsum=0;

			for(i=0,j=0;i<pvN;i+=NR,j++) {

				// Discard the non valid tests
				pvC=0; for(int k=0;k<NR;k++) if(pv[k]!=-1) pvt[pvC++]=pv[i+k];

				if(pvC>1) {
					sort_pv(pvt,pvC); OvTST=0;
					pvKS[j]=KStest(pvt,pvC);
					pvAD[j]=ADtest(pvt,pvC);
					pvSL[j]=SLtest(pvt,pvC); SLsum+=pvSL[j];
					sprintf(buf,"%d\t%f\t%f\t%7.3f\n",1+i/NR,pvKS[j],pvAD[j],pvSL[j]);
					PRT(buf);
				}
			}

			sort_pv(pvKS,NT); sort_pv(pvAD,NT);
			sprintf(buf,"\nOverall\t%f\t%f\t%8.4f\n\n",
				KStest(pvKS,NT),ADtest(pvAD,NT),SLsum/NT); PRT(buf);
		}

		// Overall testNAME p-value
		// ------------------------

		sprintf(buf,"pvals\tTEST name\tKS\tAD\tSL\n"); PRT(buf); SLsum=0;

		OvTST=0; ULONG *TSTid=new ULONG[NR];

		for(i=0;i<NR;i++) {
			pvC=0; for(int j=0;j<NT;j++) if(pv[i+j*NR]!=-1) pvt[pvC++]=pv[i+j*NR];
			if(pvC>1) { // Tests with all -1's
				OvTST++; // Valid overall p-values
				sort_pv(pvt,pvC);
				pvKS[i]=KStest(pvt,pvC);
				pvAD[i]=ADtest(pvt,pvC);
				pvSL[i]=SLtest(pvt,pvC); SLsum+=pvSL[i];
			}
			else { pvKS[i]=-1; pvAD[i]=-1; pvSL[i]=100; }
			sprintf(buf,"%5d  %*s\t%f\t%f\t%7.3f\n",pvC,maxlen,RowName[i],pvKS[i],pvAD[i],pvSL[i]); PRT(buf);
		}

		if(OvTST>1) { // Overall is meaningless for just one test

			// Sorted by KS
			if(sortKS) {
				for(i=0;i<NR;i++) { pvt[i]=pvKS[i]; TSTid[i]=i; }
				SORT(pvt,TSTid,0,NR-1,0); strcpy(buf,"\n\t= Sorted by KS =\n"); PRT(buf);
				for(i=0;i<NR;i++) {
					sprintf(buf,"%*s\t%f\t%f\t%7.3f\n",maxlen,RowName[TSTid[i]],pvt[i],pvAD[TSTid[i]],pvSL[TSTid[i]]);
					PRT(buf);
				}
			}

			// Sorted by AD
			if(sortAD) {
				for(i=0;i<NR;i++) { pvt[i]=pvAD[i]; TSTid[i]=i; }
				SORT(pvt,TSTid,0,NR-1,0); strcpy(buf,"\n\t= Sorted by AD =\n"); PRT(buf);
				for(i=0;i<NR;i++) {
					sprintf(buf,"%*s\t%f\t%f\t%7.3f\n",maxlen,RowName[TSTid[i]],pvKS[TSTid[i]],pvt[i],pvSL[TSTid[i]]);
					PRT(buf);
				}
			}

			// Sorted by SL
			if(sortSL) {
				for(i=0;i<NR;i++) { pvt[i]=pvSL[i]; TSTid[i]=i; }
				SORT(pvt,TSTid,0,NR-1,1); strcpy(buf,"\n\t= Sorted by SL =\n"); PRT(buf);
				for(i=0;i<NR;i++) {
					sprintf(buf,"%*s\t%f\t%f\t%7.3f\n",maxlen,RowName[TSTid[i]],pvKS[TSTid[i]],pvAD[TSTid[i]],pvt[i]);
					PRT(buf);
				}
			}

			sort_pv(pvKS,NR); sort_pv(pvAD,NR);
			sprintf(buf,"\n%*s\t%f\t%f\t%8.4f\n\n",maxlen,"Overall",
				KStest(pvKS,NR),ADtest(pvAD,NR),SLsum/NR); PRT(buf);
		}
		else {
			if(OvTST==0 && NR>1) PRT(" !! All the tests have failed !!\n");
			else PRT("\n");
		}

		delete[] pvKS; delete[] pvAD; delete[] pvSL; delete[] TSTid;
		for(i=0;i<NR;i++) delete[] RowName[i]; delete[] RowName;

	} // if(NT>2) (NT>1)

	// P-value of all the p-values
	// ---------------------------

	pvC=0; for(i=0;i<pvN;i++) if(pv[i]!=-1) pvt[pvC++]=pv[i];
	if(pvC>1) {
		sprintf(buf,"All the %d p-values\tKS\tAD\tSL\n",pvC); PRT(buf);
		sort_pv(pvt,pvC);
		sprintf(buf,"Overall\t%f\t%f\t%7.3f\n",
			KStest(pvt,pvC),ADtest(pvt,pvC),SLtest(pvt,pvC)); PRT(buf);
	}

	if(NT) {
			#ifdef USE_PERF_COUNTER // Running time
		QueryPerformanceCounter(&tn); double T=(tn.QuadPart-RunTime.QuadPart)/double(freq.QuadPart);
			#else
		double T=(clock()-RunTime)/(double)CLOCKS_PER_SEC;
			#endif
		if(NT!=NTorig) { sprintf(buf,"\n\t!! WARNING !!   Only %d sequence(s) have been tested.\n\n",NT); PRT(buf); }
		sprintf(buf,"\nRunning time: %s\n",TIMEtoHMS(buf+90,T,2));
		PRT(buf); printf("\nResults saved in: \"%s\"\n",OutFile);
	}
	else puts(ERRO"No sequence tested.\n");

	delete[] pv; delete[] pvt; delete[] bits; DEL;

	#if 0 // BEEP
	puts("\n\tPress any key to exit");
BEEP: for(int b=0;b<2;b++) { printf("\a"); _sleep(200); }
	for(b=0;b<300;b++) {
		if(kbhit()) return 0;
		_sleep(10);
	} goto BEEP;
	#else
	return 0;
	#endif
}
