/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

#ifdef __GNUC__
#include <sys/types.h>
#endif

#ifdef _PROVA_

//static UI64 _FASTCALL_ RDTSC(void) // Used only for debug
UI64 _FASTCALL_ RDTSC(void) // Used only for debug
{
	UI64 TS,*pTS=&TS;
	_asm {
		//cpuid // force all previous instructions to complete
		rdtsc
		mov ebx, pTS // Put the address where the timestamp goes in ebx.
		mov [ebx],eax   // Save low 4 bytes of timestamp there.
		mov [ebx+4],edx // Save high 4 bytes of timestamp next.
	}
	return TS;
}

ULONG _FASTCALL_ init(void) { return (ULONG)RDTSC(); }
ULONG MT19937(void);

#endif // _PROVA_


/*
	===============================================================================
	THIS GENERATOR IS ONLY NEEDED TO TEST THE DH TESTS (writing "RaBiGeTe DHcheck")
	===============================================================================
*/

// Maximally-Equidistributed Collision-Free combined LFSR
ULONG _FASTCALL_ lfsr113(void) // Period ~2^113 ~10^34
{
	static ULONG z1,z2,z3,z4,RISE=1; ULONG b;
	if(RISE) {
		RISE=0;
			#ifdef _PROVA_
		do z1=init(); while(z1<2);
		do z2=init(); while(z2<8);
		do z3=init(); while(z3<16);
		do z4=init(); while(z4<128);
		//z1=z2=z3=z4=9999; puts("\n\t\t -- TEST VECTOR --\n");
			#else
		z1=z2=z3=z4=9999; puts("\n\t\t -- TEST VECTOR --\n");
			#endif
	}
	b=((z1<< 6)^z1) >> 13; z1=((z1 & 4294967294UL) << 18) ^ b;
	b=((z2<< 2)^z2) >> 27; z2=((z2 & 4294967288UL) <<  2) ^ b;
	b=((z3<<13)^z3) >> 21; z3=((z3 & 4294967280UL) <<  7) ^ b;
	b=((z4<< 3)^z4) >> 12; z4=((z4 & 4294967168UL) << 13) ^ b;
	return z1^z2^z3^z4; // Same output for little/big endian
}

//---------------------------------------------------------------------------

#ifdef _PROVA_

// RC6

#define RC6b 32 // Key length [bytes]

void rc6_setup(unsigned char *K, int b);
void rc6_encrypt(unsigned long *pt, unsigned long *ct);

static ULONG _FASTCALL_ RC6rnd(void)
{
	static ULONG buf[4]; static int j=0,flg=1,C=init();
	if(flg) { // Initialization
		flg=0; UCHAR key[RC6b]; 
		for(int j=0;j<RC6b;j++) key[j]=lfsr113()>>24;
		rc6_setup(key,RC6b);
	}
	if(j==0) {
		for(int i=0;i<4;i++) buf[i]=MT19937();//C++
		rc6_encrypt(buf,buf);
		j=4; //buf[0]^=buf[2]; buf[1]^=buf[3]; // Compression
	}
	return buf[--j];
}

static int _FASTCALL_ RC6bit(void)
{
	static ULONG r,C=0;
	if(C==0) { r=RC6rnd(); C=32; }
	int b=r&1; r>>=1; C--; return b;
}


//---------------------------------------------------------------------------

// AES

#define AESb 128 // Key length [bits]

#include "rijndael-alg-fst.h"

static ULONG _FASTCALL_ AESrnd(void)
{
	static ULONG buf[4]; static int j=0,flg=1,Nr;
	if(flg) { // Initialization
		flg=0; UCHAR key[AESb/8]; 
		for(int j=0;j<AESb/8;j++) key[j]=lfsr113()>>24;
		Nr=rijndaelKeySetupEnc(key,AESb);
	}
	if(j==0) {
		for(int i=0;i<4;i++) buf[i]=MT19937();
		rijndaelEncrypt(Nr,(UCHAR*)buf,(UCHAR*)buf); j=4;
	}
	return buf[--j];
}

static int _FASTCALL_ AESbit(void)
{
	static ULONG r,C=0;
	if(C==0) { r=AESrnd(); C=32; }
	int b=r&1; r>>=1; C--; return b;
}


//---------------------------------------------------------------------------

static ULONG _FASTCALL_ LCG(void)
{
	static ULONG y=lfsr113();

	y=y*2147001325L+715136305L; // genacpc.ps: LCG(2^32,2147001325,715136305,0)
	//y=y*69069L;
	//y=(y*1099087573ui64)%0x100000000ui64;
	//y=(UI64(y)*1588635695ui64)%4294967291ui64;
	//y=UI64(y)*2891336453ui64+1ui64;
	return y;
}

// LFSR (period: 2^bit -1)
static int _FASTCALL_ LFSR(void)
{
	static UI64 s,poly; static int ini=1;
	if(ini) { ini=0; 
		#define bit 23
		if(bit==14) poly=0x3ff3;		//        16 k
		if(bit==16) poly=0xf5cf;		//        64 k
		if(bit==18) poly=0x3f618;		//       256 k
		if(bit==19) poly=0x67fae;		//       512 k
		if(bit==20) poly=0xfbb90;		//     1 024 k
		if(bit==22) poly=0x3acccf;		//     4 096 k
		if(bit==23) poly=0x57cf80;		//     8 192 k
		if(bit==24) poly=0xfe5e00;		//    16 384 k
		if(bit==25) poly=0x1f7fd62;		//    32 768 k
		if(bit==26) poly=0x357ef66;		//    65 535 k
		if(bit==27) poly=0x7f3ffd6;		//   131 072 k
		if(bit==28) poly=0xbb5f22f;		//   262 144 k
		if(bit==29) poly=0x1feefeb8;	//   524 288 k
		if(bit==30) poly=0x37fec8ec;	// 1 048 576 k
		if(bit==31) poly=0x66e7faa3;	// 2 097 152 k
		if(bit==32) poly=0xfb9e5ddf;	// 4 194 304 k
//		if(bit==33) poly=0x1fafd7c8cULL;// 8 388 608 k
		if(bit==33) poly=0x1fafd7c8cui64;// 8 388 608 k
		do s=( UI64(RC6rnd())<<32 | RC6rnd() ) & ((UI64(1)<<bit)-UI64(1)); while(s==0);
	}
	if(s & 1) { s=(s>>1)^poly; return 1; }
	s>>=1; return 0;
}

#define GENE() LFSR()
static int _FASTCALL_ MLS(void)
{
	static ULONG ones=0,zeros=0; ULONG mask; int b;
	do {
		GENE() ? ++ones : ++zeros;
		(b = GENE()) ? ++ones : ++zeros;
	} while(!((mask = ones & zeros)));
	ones -= mask; zeros -= mask;
	return b;
}

static ULONG _FASTCALL_ RCXrnd(void) // Based on RC5/RC6 encrypt function
{
		#if 1 // RC5 based

	ULONG A=RDTSC()^RDTSC(), B=RDTSC(); // AMLS fails ("i<5")
	//ULONG A=RDTSC(), B=3141592621UL*RDTSC();
	//ULONG A=RDTSC(), B=RDTSC();
int R=3; //static int C=0; if(++C==16) { R=2; C=0; }
	for(int i=0; i<R; i++) { A = _lrotr(A^B,B); B = _lrotr(B^A,A); }
	return A^B;

		#else // RC6 based
	ULONG A=RDTSC(), B=RDTSC(),C=RDTSC(),D=RDTSC();
	for(int i=0; i<3; i++) {
		ULONG t = _lrotr(B * (2 * B + 1), 5), u = _lrotr(D * (2 * D + 1), 5);
		A = _lrotr(A ^ t, u); C = _lrotr(C ^ u, t);
		ULONG x = A; A = B; B = C; C = D; D = x;
	}
	return A;//^B;//^C^D;
		#endif
}


void RBG(void)
{
#if 1 // ULONG generators
	for(ULONG i=0;i<=n>>5;i++) {
		//bits[i]=LCG();
		//bits[i]=lfsr113();
		//bits[i]=MT19937();
		//bits[i]=CMWC4096();
		//bits[i]=RC6rnd();
		bits[i]=AESrnd();
		//bits[i]=RCXrnd();
		//bits[i]=3141592621UL*RDTSC();
		//bits[i]=(RC6rnd()>>26) ^ RDTSC()*RDTSC();
		//bits[i]=RDTSC()*RDTSC();
		//bits[i]=(RC6rnd()>>20) ^ LCG();
		//bits[i]=(lfsr113()>>26) ^ LCG();
		//bits[i]=_lrotr( RDTSC()*RDTSC() >>3 ,RDTSC()>>5 )*3141592621ul;
		//bits[i]=_lrotr( (RDTSC()*3141592621ul)>>1 ,RDTSC()>>1 );
		// Invert the bits
//ULONG x,b=bits[i]; for(int j=0;j<32;j++) { x=x<<1 | b&1; b>>=1; } bits[i]=x;
		#if 0 // MT19937: check how fast the diffusion is
		void init_by_array(void); init_by_array();
		for(int j=0;j<624*100;j++) MT19937(); // Diffusion
		for(int j=0;j<624;j++) bits[i++]=MT19937(); i--; // Get the whole state
		#endif

		/*/ Interleaved (In b bits there are b-1 bad bits)
		static ULONG f=1,m; if(f) { f=0; m=1; int b=2; for(int i=0;i<32;i+=b) m=m<<b|1; }
		bits[i]=RC6rnd()&m | LCG()&(m^0xffffffff);//*/
	}

#else // Bit generators

	for(ULONG i=0;i<n;i++) {
		int z;
			#if 0 // Interleaved BAD (1 GOOD every X BAD)
		//static int b=0; if(++b==3) { b=0; z=AESbit(); } else
		static int b=0; if(0==b--) z=AESbit(); else { b=(lfsr113()&3)+2;
			//z=LFSR();
			//do z=LFSR(); while(LFSR());
			//z=(LCG()>>26) & 1;
			//z=(RDTSC()*RDTSC()>>16) &1;
			z=MLS();
		}
			#elif 0 // Interleaved GOOD (1 BAD every X GOOD)
		//static int b=0; if(++b<=2) { b=0; z=RC6bit(); } else
		static int b=0; if(b-->0) z=AESbit(); else { b=(lfsr113()&7)+3;
			//do z=LFSR(); while(LFSR());
			//z=MLS();
			//z=LFSR();
			//z=(LCG()>>7) & 1;
			//z=(RDTSC()*RDTSC()>>8) &1;
		}
			#else
		//switch(i%7) {
		//switch(lfsr113()%7) {
		switch(1) {
		case 0: z=RC6bit(); break;
		case 1: do z=LFSR(); while(LFSR()); break;
		case 2: z=LFSR(); break;
		case 3: z=MLS(); break;
		case 4: z=(LCG()>>12) & 1; break; // Test a single bit of a 32-bit generator
		case 5: z=(RDTSC()*RDTSC()>>10) &1; break;
		default: z=(lfsr113()>>1) &1; // k=0 in gorilla = lfsr113()>>31
		}
			#endif

		set_bit(i,z,bits);
	}
#endif

/*	for(int i=0;i<n>>5;i++) { // Little endian -> big endian
		ULONG tmp = bits[i]; tmp = (tmp >> 16) | (tmp << 16);
		bits[i] = ((tmp & 0xff00ff00UL) >> 8) | ((tmp & 0x00ff00ffUL) << 8);
	}//*/
}

#else // _PROVA_


/*
	==================================================================
	THIS GENERATES THE SEQUENCE TO TEST USING THE HARD CODED GENERATOR
	==================================================================
*/
void RBG(void)
{
	for(int i=0;i<n>>5;i++) bits[i]=lfsr113(); // <-- REPLACE lfsr113 WITH THE GENERATOR TO TEST
}

#endif // _PROVA_
