/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

extern int AMLSmin,AMLSmax,AMLSstp;

#define TEST(x,a,b)  ( (x[a]>>(b)) & 1 )
#define SET1(x,a,b)  (x[a] |= (1uL<<(b)))
#define SET0(x,a,b)  (x[a] &=~(1uL<<(b)))
#define SET(x,a,b,c) ((c)?SET1(x,a,b):SET0(x,a,b))
#define FLP(x,a,b)   (x[a] ^= (1uL<<(b)))
#define FLIP(x,a,b)  ((FLP(x,a,b)>>(b))&1)

static ULONG bcount,*exists,*values,BLK_LEN;

// The AMLS test is about 13% faster with __fastcall

static void _FASTCALL_ flipped(int node, int coin)
{
	while( node < BLK_LEN ) { // MaxLevel = log2(BLK_LEN)
		const int a = node>>5, b = node&31;
		if( FLIP(exists,a,b) ) { SET(values,a,b,coin); return; }
		node <<= 1;
		if( TEST(values,a,b) == coin ) { flipped(node|1, coin); coin = 0; }
		else { ++bcount; coin = 1; }
	}
}

static int _FASTCALL_ amls(const int i)
{
	int c=0,BITSc=0; const ULONG *bts=bits+BLK_LEN/32*i;
	memset(exists,0,BLK_LEN/8); 
	do {
		bcount = 0;
		do {
			SET1(exists,0,1);
			SET(values,0,1,TEST(bts,BITSc>>5, BITSc&31)); BITSc++;
			flipped(1,TEST(bts,BITSc>>5, BITSc&31)); BITSc++;
		} while(!bcount);
		c+=bcount;
	} while(BITSc<BLK_LEN);
	return c; // Exactly BLK_LEN bits used
}

void AMLS(void)
{
	exists=new ULONG[AMLSmax*1024/32]; values=new ULONG[AMLSmax*1024/32];

	for(int blk=AMLSmin; blk<=AMLSmax; INCR(blk,AMLSstp))
	{
		double Ec,VARc; BLK_LEN=blk*1024; const int len=n/BLK_LEN; if(len<5) break;

		switch(blk) { // Expected mean and variance (CI @ 99.9%)

case     2: Ec= 1820.68207; // 1820.67731 < Ec < 1820.68684  (5.24e-006)
            VARc= 98.673; // 98.6097 < VARc < 98.7363  (0.0013)
            break; // nn= 52617712 (100.4 Gb)

case     4: Ec= 3725.94031; // 3725.93131 < Ec < 3725.94931  (4.83e-006)
            VARc= 159.954; // 159.802 < VARc < 160.106  (0.0019)
            break; // nn= 23920000 (91.2 Gb)

case     8: Ec= 7590.66628; // 7590.64868 < Ec < 7590.68388  (4.64e-006)
            VARc= 257.976; // 257.598 < VARc < 258.354  (0.0029)
            break; // nn= 10092000 (77.0 Gb)

case    16: Ec= 15408.1669; // 15408.1366 < Ec < 15408.1971  (3.93e-006)
            VARc= 415.276; // 414.453 < VARc < 416.101  (0.0040)
            break; // nn= 5500000 (83.9 Gb)

case    32: Ec= 31185.9255; // 31185.8795 < Ec < 31185.9715  (2.95e-006)
            VARc= 668.317; // 666.729 < VARc < 669.91  (0.0048)
            break; // nn= 3824000 (116.7 Gb)

case    64: Ec= 62972.6764; // 62972.6266 < Ec < 62972.7262  (1.58e-006)
            VARc= 1074.79; // 1072.61 < VARc < 1076.98  (0.0041)
            break; // nn= 5255200 (320.8 Gb)

case   128: Ec= 126920.711; // 126920.641 < Ec < 126920.782  (1.11e-006)
            VARc= 1732.77; // 1728.85 < VARc < 1736.7  (0.0045)
            break; // nn= 4218800 (515.0 Gb)

case   256: Ec= 255422.956; // 255422.868 < Ec < 255423.043  (6.85e-007)
            VARc= 2793.34; // 2787.17 < VARc < 2799.53  (0.0044)
            break; // nn= 4427900 (1081.0 Gb)

case   512: Ec= 513408.741; // 513408.548 < Ec < 513408.933  (7.49e-007)
            VARc= 4512.02; // 4494.8 < VARc < 4529.32  (0.0076)
            break; // nn= 1480500 (722.9 Gb)

case  1024: Ec= 1030968.28; // 1030967.76 < Ec < 1030968.8  (1.01e-006)
            VARc= 7314.56; // 7255.64 < VARc < 7374.15  (0.0162)
            break; // nn= 330000 (322.3 Gb)

case  2048: Ec= 2068657.39; // 2068656.3 < Ec < 2068658.48  (1.05e-006)
            VARc= 11757.3; // 11600.7 < VARc < 11916.9  (0.0269)
            break; // nn= 119780 (233.9 Gb)

case  4096: Ec= 4148192.47; // 4148190.99 < Ec < 4148193.95  (7.13e-007)
            VARc= 19002; // 18732.1 < VARc < 19277.3  (0.0287)
            break; // nn= 105216 (411.0 Gb)

case  8192: Ec= 8313993.87; // 8313992.23 < Ec < 8313995.52  (3.97e-007)
            VARc= 30841.8; // 30458.1 < VARc < 31232.2  (0.0251)
            break; // nn= 137530 (1074.5 Gb)

case 16384: Ec= 16656480.4; // 16656478.3 < Ec < 16656482.6  (2.56e-007)
            VARc= 50062.8; // 49429.5 < VARc < 50707.5  (0.0255)
            break; // nn= 132930 (2077.0 Gb)

case 32768: Ec= 33359073.3; // 33359070 < Ec < 33359076.5  (1.95e-007)
            VARc= 80781.4; // 79561.1 < VARc < 82028.1  (0.0305)
            break; // nn= 92892 (2902.9 Gb)
		}

		double XX=0;
		for(int i=0; i<len; i++) {
			double c=amls(i)-Ec; XX+=c*c/VARc;
		}
		sprintf(buf, "AMLS %d_%d",blk,len); SAVE(buf,chiQ(XX,len));

	} // for blk

	delete[] exists; delete[] values;
}
