/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

extern int
	AUTWdmin,AUTWdmax,AUTWdstp,
	AUTWmmin,AUTWmmax,AUTWmstp;

// Windowed autocorrelation (== autocorrelation when m==1)

extern double *LOGf; // ln x!
double _FASTCALL_ Cprob(int n,int i); // C(n,i) / 2^n

void AUTOw(void)
{
	ULONG *const s=new ULONG[AUTWmmax+1];

	for(int m = AUTWmmin; m <= AUTWmmax; INCR(m,AUTWmstp)) {

		if(m==1) { // Autocorrelation (faster)
			for(int i = AUTWdmin; i <= AUTWdmax; INCR(i,AUTWdstp)) {
				int Ad=0; for(ULONG j=0;j<n-i;j++) Ad+=get_bit(j,bits)^get_bit(j+i,bits);
				double XX=2.*(Ad-(n-i)/2.)/sqrt((double)(n-i));
				sprintf(buf,"AutoW 1_%d",i); SAVE(buf,chiQ(XX*XX,1)); // 2-tail normal distribution
			}
			continue;
		}

		const int N=n/m; int i,j;

		for(i = AUTWdmin; i <= AUTWdmax; INCR(i,AUTWdstp)) {

			if(i>N || Cprob(m,m/2)*(N-i-1)<5) { delete[] s; return; }

			for(j=0;j<=m;j++) s[j]=0;

			for(j = i; j < N; j++) {
				int b=0; // Counter for the different bits in a block
				for(int k=0;k<m;k++)
					b+=get_bit(j*m+k,bits) ^ get_bit((j-i)*m+k,bits);
				s[b]++;
			}

			double XX=0,Cp=0; int DF=0,Cn=0;
			// Grouping the first half
			for(j = 0; j <= m/2; j++) {
				Cp+=Cprob(m,j)*(N-i-1); Cn+=s[j]; if(Cp<5) continue;
				XX+=(Cn-Cp)*(Cn-Cp)/Cp; DF++;
				Cp=0; Cn=0;
			}

			// Grouping the second half (from top to mid)
			for(j = m; j > m/2; j--) {
				Cp+=Cprob(m,j)*(N-i-1); Cn+=s[j]; if(Cp<5) continue;
				XX+=(Cn-Cp)*(Cn-Cp)/Cp; DF++;
				Cp=0; Cn=0;
			}

			sprintf(buf,"AutoW %d_%d",m,i); SAVE(buf,chiQ(XX,DF-1));
		}
	} // for m

	delete[] s;
}
