/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

/*
       This is the "tough" BIRTHDAY SPACINGS TEST
Choose 4096 birthdays in a "year" of 2^32 days. Thus each    
birthday is a 32-bit integer and the test uses 2^12 of them, 
so that j, the number of duplicate spacings, is asympotically
Poisson distributed with lambda=4 .  Generators that pass the
earlier tests for m=1024 and n=2^24 often fail this test, yet
those that pass this test seem to pass the "weaker" test.
Each set of 4096 birthdays provide a Poisson variate j, and  
500 such j's lead to a chisquare test to see if the result   
is consistent with the Poisson distribution with lambda=16.  
*/

#include "defi.h"
#include <stdlib.h>

#ifdef __GNUC__
int compi(const void *i, const void *j) //compare function for qsort
#else
int __cdecl compi(const void *i, const void *j) //compare function for qsort
#endif
{
	if ( *(ULONG*)i < *(ULONG*)j ) return -1;
	if ( *(ULONG*)i > *(ULONG*)j ) return 1;
	return 0;
}

void BDAY(void)
{
	if (n < 2048000*32) return;
	const int m=4096; int i,k;
    ULONG *const t=new ULONG[4096],obs[11]={0,0,0,0,0,0,0,0,0,0,0};
    double w,x,ex[11],v=0.; bpos=0;

	for(k=1; k<=500; k++) { //loop for sample of size 500
		for(i=0; i<m; i++) t[i]=rng(); // get 4096 IUNI's
		qsort(t,m,sizeof(long),compi); // sort them

		for(i=m-1; i>0; i--) t[i]=t[i]-t[i-1]; // get spacings
		qsort(t+1,m-1,sizeof(long),compi); // sort spacings

		int j=0; for(i=1; i<m-1; i++) j+=(t[i]==t[i+1]);
		if(j>10) j=10;
		obs[j]++; // count duplicate spacings
	}// end  sample of 500 loop

	ex[0]=500.*exp(-4.); ex[10]=500.*.008132242;
    for(i=1;i<10;i++) ex[i]=4.*ex[i-1]/i;
    for(i=0;i<11;i++) v+= (obs[i]-ex[i])*(obs[i]-ex[i])/ex[i];

	x=.5*v; w=24.+(24.+(12.+(4.+x)*x)*x)*x;
	SAVE("Bday",1.-exp(-x)*w/24.); delete[] t;
}
