/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"
#include <string.h>
#include <stdlib.h>

extern int COUmin,COUmax,COUstp;

// Coupon collector's test  [The Art of Computer Programming PAGE 62 (76/704)]

void COUPON(void)
{
	FILE *prob=fopen("DAT_coupon_prob.txt","rb");

	for(int s=COUmin;s<=COUmax; INCR(s,COUstp)) { // Block size [bits]

		const ULONG d=1ul<<s; // Numbers from 0 to d-1

		// Find the starting position for d and find rmin and t
		while(!feof(prob)) {
			fgets(buf,999,prob);
			if(*buf=='-') if(atoi(buf+1)==d) break; // d found
		}
		if(feof(prob)) break; // d not available

		const int rmin=atoi(strchr(buf,'\t')),t=atoi(strrchr(buf,'\t'));

		// This is just a preliminary check which catch big deviations from the requirements
		if(n/double(t-rmin)/d/s/5.<.05) break; // n/s/d/(t-rmin)/5. ~ counts for each bin (5 is not related to the 5 in the chi-square test)

		ULONG b=0,*const C=new ULONG[t+1],*const flg=new ULONG[d];
		memset(C,0,(t+1)*sizeof(ULONG)); ULONG N=0;

		C0: // Collect the coupons (don't use get_bit(y,flg))
			ULONG r=0,c=0; memset(flg,0,d*sizeof(ULONG));
				
				#if 1 // FASTER (+8.7%)

			collect:
				r++; ULONG y=0; for(int j=0;j<s;j++) y=y*2 + get_bit(b++,bits);
				if(b<n-s) {
					if(flg[y]) goto collect;
					flg[y]=1; if(++c<d) goto collect;
				}

			if(c==d) { // Found a complete set of coupons using r coupons
				C[min(r,t)]++; N++; goto C0; // If r >= t we increment the count for t
			}								 // We have N complete sets
				#else // SLOWER

			while(b<n-s) {
				r++; ULONG y=0; for(int j=0;j<s;j++) y=y*2 + get_bit(b++,bits);
				if(flg[y]==0) {
					flg[y]=1;
					if(++c==d) { C[min(r,t)]++; N++; goto C0; }
				}
			}
				#endif
/*
	L = (N-C[t]) / (t-rmin) is an empirical "sensitiveness" or "reliability level"
	N is the number of the complete sets found
	t-rmin is the number of available bins
	In both the numerator and the denominator it is excluded t (which is a "cumulative" bin)
	So that L is approximately the number of coupons in each bin:
	a big L means a big "sensitiveness" or "reliability level" because there are many bins to analyze.
	But remember: it's just an empirical index.
*/
		double L=(N-C[t])/double(t-rmin);
		if(L>3.6) { // Second check: when L<3.6 the test is not reliable (few samples)
			double XX=0,Pc=0; ULONG CC=0,DF=0; //printf("Dr= %d\tE= %f\tN= %d\tE= %.1f  L= %.3f   \n",t-rmin, n/double(t-rmin)/d/s/5., N, n/s/d/1., (N-C[t])/double(t-rmin) );
			for(int i=d;i<=t;i++) {
				CC+=C[i]; // Grouping the bins
				if(i>=rmin) { // We have the probabilities for these i's
					fgets(buf,999,prob); // Probability for i
					Pc+=atof(buf); // Cumulative probabilty
					if(N*Pc>=5) { //if(CC) printf("%d\t %d\t %.1f\t %.9g\n",i,CC,N*Pc,Pc);
						XX+=double(CC)/Pc*CC; Pc=0; CC=0; DF++;
					}
				}
			}
			sprintf(buf,"Coupon %d_%.0f",s,L); SAVE(buf,chiQ(XX/N-N,DF-1));
		}

		delete[] C; delete[] flg;
	}
	fclose(prob);
}
