/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include <stdlib.h>
#include "defi.h"

// normal(x) = phi(x)
static double _FASTCALL_ normal(double x)
{
		#if 0 // ~4% faster (the error in the overall p-value is ~5e-4)

	double v = 1.0 / (1.0 + 0.33267 * (x > 0.0 ? x : -x));
	v = exp(-0.5 * x * x) * (0.1740120799 + (0.3739278012 * v - 0.04793993633) * v) * v;
	return (x > 0.0 ? 1.0 - v : v);

		#else

	//return 0.5 * (x > 0.0 ? (2.0 - erfc(x / M_SQRT2)) : erfc(-x / M_SQRT2));
	return Phi(x);

		#endif
}


void CUSUM(void)
{
	int	k, cusum; double fsum,y, z; ULONG i;
	
	k = 0; cusum = 1; fsum = 0.0; i = 0;
	do {
		k += (get_bit(i,bits) ? 1 : -1); cusum = max(cusum, abs(k));
	} while(++i < n);
	
	z = cusum / sqrt((double)n); cusum = 4 * ((n/ cusum - 1) / 4) + 1; y = -z * cusum;
	for(k = 0; k <= cusum; ++k) fsum = normal(y += z + z) - fsum;
	fsum += fsum - normal(-z * cusum) - normal(z * (cusum + 2));
	SAVE("Cusum 1", 1.0 + fsum);

	k = 0; cusum = 1; fsum = 0.0; i = n;
	while(i--) {
		k += (get_bit(i,bits) ? 1 : -1); cusum = max(cusum, abs(k));
	}
	
	z = cusum / sqrt((double)n); cusum = 4 * ((n/ cusum - 1) / 4) + 1; y = -z * cusum;
	for(k = 0; k <= cusum; ++k) fsum = normal(y += z + z) - fsum;
	fsum += fsum - normal(-z * cusum) - normal(z * (cusum + 2));
	SAVE("Cusum 2", 1.0 + fsum);
}
