/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

extern int LBLKmin,LBLKmax,LBLKstp, // Long block parameters
	SBLKsmin,SBLKsmax,SBLKsstp, // Short block (block size [bits])
	SBLKdmin,SBLKdmax,SBLKdstp; // Short block (bit distance [bits])

extern int pvN; // Number of p-values in pv
extern double *pv; // P-values obtained from the tests

/*	Short blocks (generalized variant)
	Setting blkD=1 and blkS=4 yields the FIPS140 poker test.
	Setting blkD=1 and blkS=1 yields the NIST and FIPS140 frequency (monobit) test.
	Short blk 1_1 = Serial 1_1 = Serial 1_2
*/
void SHORTBLK(void)
{
	int pvNi=pvN; // Index of the 1st p-value saved by this test

	for(int blkS=SBLKsmin; blkS<=SBLKsmax; INCR(blkS,SBLKsstp)) { // Block size [bits]

		const ULONG blkN=1ul<<blkS; // Number of possible blocks

		if(n/blkS<5*blkN) break; // At least 5 counts in each bin

		ULONG *const C=new ULONG[blkN]; // Block frequency

		for(int blkD=SBLKdmin; blkD<=SBLKdmax; INCR(blkD,SBLKdstp)) { // Bits distance [bits]
			
			if(blkD>n/blkS) break;
			memset(C,0,blkN*sizeof(ULONG)); ULONG i;

			if(blkD==1) // Adjacent bits (faster version)
				for(i=0;i<=n-blkS;i+=blkS) {
					ULONG c=0; for(ULONG j=0;j<blkS;j++) c=c*2 + get_bit(i+j,bits);
					C[c]++;
				}
			else {
				const ULONG mulN=n/blkS/blkD; // Number of (blkS*blkD)-bit blocks (multi-blocks)
				for(i=0;i<mulN;i++) {
					const ULONG b=i*blkS*blkD;
					for(ULONG k=0; k<blkD; k++) {
						ULONG c=0;
						for(ULONG j=0;j<blkS;j++) c=c*2 + get_bit(b+k+blkD*j,bits);
						C[c]++;
					}
				}
			}

			double XX=0,E,S=0; for(i=0;i<blkN;i++) S+=C[i];
			E=S/blkN; // Expected
			for(i=0;i<blkN;i++) XX+=(C[i]-E)*(C[i]-E)/E;
			double pval=chiQ(XX,blkN-1);

			if(blkS==1) // Discards the duplicate p-values
				if(pvNi<pvN) { // The 1st p-value is always saved
					int p; for(p=pvNi;p<pvN;p++) if(pv[p]==pval) break;
					if(p<pvN) continue;
				}

			sprintf(buf,"Short blk %d_%d",blkS,blkD); SAVE(buf,pval);

		} // blkD

		delete[] C;

	} // blkS
}


void LONGBLK(void) // NIST's BlockFrequency
{
	for(int m=LBLKmin; m<=LBLKmax; INCR(m,LBLKstp)) {

		double sum=0;
		const ULONG N = n/m; // # OF SUBSTRING BLOCKS

		for(ULONG i = 0; i < N; i++) { // FOR EACH SUBSTRING BLOCK
			ULONG blockSum=0; // COMPUTE the i-th Pi Value
			for(int j = 0; j < m; j++) blockSum +=get_bit(j+i*m,bits);
			double pi = blockSum/(double)m -.5; sum += pi*pi;
		}
		sprintf(buf,"Long blk %d",m); SAVE(buf,chiQ(4.*m*sum,N));
	}
}


void TWOBIT(void) // two-bit test (overlapping)
{
	int n0,n1,n00,n01,n10,n11; n0=n1=n00=n01=n10=n11=0;
	int b0=get_bit(0,bits);

	for(ULONG i=1;i<n;i++) {
		int b1=get_bit(i,bits);
		if(b0==0) {
			n0++; if(b1==0) n00++; else n01++;
		}
		else {
			n1++; if(b1==0) n10++; else n11++;
		}
		b0=b1;
	}
	if(get_bit(n-1,bits)==0) n0++; else n1++; // Don't use b0 here (slower)

	double XX=4./(n-1.)*(double(n00)*n00+double(n01)*n01+double(n10)*n10+double(n11)*n11)-2./n*(double(n0)*n0+double(n1)*n1)+1.;
	SAVE("Two-bit",chiQ(XX,2));
}



static double _FASTCALL_ EXP(const int r) { return (n-r+3)/double(1ul<< r+2); }

static double _FASTCALL_ PVAL(const int BIT,ULONG *const C,const ULONG Kruns)
{
	const int NoBIT=1-BIT; memset(C+1,0,Kruns*sizeof(ULONG)); ULONG p,b=0;
	do {
		p=b; while(get_bit(p,bits)==NoBIT) if(++p>=n-1) goto endC; // Discard unwanted bits
		b=p; while(get_bit(++b,bits)==BIT) if(b>=n-1) goto endC; // Calculate the length
		ULONG l=b-p; if(l>Kruns) l=Kruns;
		C[l]++;
	} while(b<n);
	endC:

	double XX=0; // Calculate X only for Kruns with e>=5
	for(int i=1;i<=Kruns;i++) {
		double e; if(i==Kruns) e=EXP(i-1); else e=EXP(i); //printf("%d  %d  %f\n",i,C[i],e);
		XX+=(C[i]-e)*(C[i]-e)/e;
	}
	return chiQ(XX,Kruns-1);
}

void BLKGAP(void)
{
	ULONG Kruns=1; while(EXP(++Kruns)>=5);
	ULONG *const C=new ULONG[Kruns+1]; // Counter for blocks and gaps from 1 to Kruns

	SAVE("BlkGap BLK",PVAL(1,C,Kruns)); // Blocks (111...)
	SAVE("BlkGap GAP",PVAL(0,C,Kruns)); // Gaps (000...)
	delete[] C;
}
