/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

/*
This is the GCD TEST.   Let the (32-bit) RNG produce two
successive integers u,v.  Use Euclids algorithm to find the
gcd, say x, of u and v. Let k be the number of steps needed
to get x.   Then k is approximately binomial with p=.376
and n=50,  while the distribution of x is very close to
Pr(x=i)=c/i^2, with c=6/pi^2.   The gcd test uses ten
million such pairs u,v to see if the resulting frequencies
of k's and x's are consistent with the above distributions.
Congruential RNG's---even those with prime modulus---fail
this test for the distribution of k, the number of steps,
and often for the distribution of gcd values x as well.
*/

#include "defi.h"

#define nn 10000000 // pairs of IUNI's

void GCD(void)
{ 
	bpos=0; ULONG N=n/32; if(N<20000000) return;
	ULONG u, v, w, i, k, t[47], gcd[101];
	ULONG kp[36]=
		{0,0,0,5506,29532,144541,590691,2065333,6277273,16797273,39964829,
		85160313,163520964,284315128,449367802,647663189,853365977,1030017639,
		1140689999,1160424679,1085307890,933657875,738971259,538010076,360199303,
		221583256,125137422,64787412,30697488,13285746,5238570,1876493,608568,
		177920,46632,13674};

	for(i=1;i<101;i++) { gcd[i]=0; if(i<47) t[i]=0; }

	for (i=1; i<=nn; i++) {
		do { u=rng(); N--; } while(u==0);
		do { v=rng(); N--; } while(v==0);
		if(N+i*2<20000000) { puts("\n\tGCD: too many 0's in the input sequence\n"); break; }

		k=0; do { w=u%v; u=v; v=w; k++; } while(v>0);

		if (u>100) u=100;
		gcd[u]++;
		if (k<3) k=3;
		if (k>35) k=35;
		t[k]++;
	}

	double ch32=0,e;
	// Euclid's algorithm
	for (i=3; i<=35; i++) {
		e=(1e-10*kp[i])*nn; ch32+=(t[i]-e)*(t[i]-e)/e;
	}

	SAVE("GCD stp",chiQ(ch32,32)); // steps to gcd

	e=nn*.61097691e-2; double ch99=(gcd[100]-e)*(gcd[100]-e)/e;
	for (i=1; i<100; i++) {
		e=nn*0.6079271/(i*i); ch99+=(gcd[i]-e)*(gcd[i]-e)/e;
	}

	SAVE("GCD dst",chiQ(ch99,99)); // dist. of gcd's
}

