/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

extern int MAUmin,MAUmax,MAUstp,MAU_n;

// Maurers universal statistical test
// [it01.ps] An Accurate Evaluation of Maurer's Universal Test
// [Statistical_test.ps] A Universal Statistical Test for Random Bit Generators (page 12)
// HAC ( 5.4.5 page 183)

static double MAURER(const int L,const ULONG N)
{
	struct MeVar { double m,v,d,e; };
	MeVar const mv[]={
	//             var.
		0.7326495,0.6897678,.0      ,.0      , // L=1 (not used)
		1.5374383,1.3377388,.2363135,.5594976, // 2
		2.4016068,1.9013347,.2732725,.4890883, // 3
		3.3112247,2.3577369,.3045101,.4435381, // 4
		4.2534266,2.7045528,.3296587,.4137196, // 5
		5.2177052,2.9540324,.3489769,.3941338, // 6
		6.1962507,3.1253919,.3631815,.3813210, // 7
		7.1836656,3.2386622,.3732189,.3730195, // 8
		8.1764248,3.3112009,.3800637,.3677118, // 9
		9.1723243,3.3564569,.3845867,.3643695, // 10
		10.170032,3.3840870,.3874942,.3622979, // 11
		11.168765,3.4006541,.3893189,.3610336, // 12
		12.168070,3.4104380,.3904405,.3602731, // 13
		13.167693,3.4161418,.3911178,.3598216, // 14
		14.167488,3.4194304,.3915202,.3595571, // 15
		15.167379,3.4213083,.3917561,.3594040  // L=16
	};

	ULONG *const T=new ULONG[1ul<<L],
		*const flg=new ULONG[(1ul<<L)/32+1]; // bit =1: number generated

	ULONG flgC=1ul<<L; // Counter for the number bi still not generated
	memset(flg,0,((1ul<<L)/32+1)*sizeof(ULONG));

	// Initialize T with the last position of the L-bit blocks
	ULONG Q=0;
	do {
		ULONG x=0; for(int j=0;j<L;j++) x=x*2 + get_bit(Q*L+j,bits);
		T[x]=Q; if(get_bit(x,flg)==0) { set_bit(x,1,flg); --flgC; }
		if(++Q==N) { 
			puts("\tMAURER: sequence too bad!"); 
			delete[] flg; delete[] T; return -1;
		}
	} while(flgC);

	const int K=N-Q; delete[] flg;
	if(K<33*(1ul<<L)) {
		puts("\n\tMAURER: sequence too short.\n"); delete[] T; return -1;
	}

	double sum=0;
	for(ULONG i=Q;i<N;i++) {
		ULONG x=0; for(int j=0;j<L;j++) x=x*2 + get_bit(i*L+j,bits);
		sum+=log((double)(i-T[x])); T[x]=i;
	}

	delete[] T; const double fTu=sum/K/M_LN2;

	// [it01.ps] For K >= 332^L we get with an excellent approximation:
	double sig=sqrt((mv[L-1].d+mv[L-1].e*(1ul<<L)/K)*mv[L-1].v/K);

	double xx=(fTu-mv[L-1].m)/sig;
	return chiQ(xx*xx,1); // Two-sided
}


void MAURER(void)
{
	for(int L=MAUmin; L<=MAUmax; INCR(L,MAUstp)) {
		const ULONG N=MAU_n*(1ul<<L); if(n<N*L) break;
		double pv=MAURER(L,N);
		sprintf(buf,"Maurer %d",L); SAVE(buf,pv);
	}
}
