/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

extern int NOTMmin,NOTMmax,NOTMstp,NOTMtemp;

#define Bb 32
static int *templ; // m-bit template to test
static ULONG numOfTemplates, // Templates to test
	SKIP, // Templates to skip
	skip, // Templates skipped
	V,A[Bb];
static const ULONG displayMask = 1ul << (Bb-1);

void CreateTemplate(const int m)
{
	int match,b,c;
	do {
		V++; ULONG i,v=V;

		for(c = 0; c < Bb; c++) {
			if (v & displayMask) A[c] = 1; else A[c] = 0;
			v <<= 1;
		}

		for(i = 1; i < m; i++) {
			match = 1;
			if ((A[Bb-m]!= A[Bb-1]) && ((A[Bb-m]!= A[Bb-2])||(A[Bb-m+1] != A[Bb-1])))
				for(c = Bb-m; c <= (Bb-1)-i; c++)
					if (A[c] != A[c+i]) { match = 0; break; }
			if(match) break;
		}
	} while(match || ++skip<SKIP);

	skip=0; for(b=0,c=Bb-m;c<Bb;c++,b++) templ[b]=A[c];
}


void NOTM(void)
{
	ULONG n_templ[]={ // Number of templates for the m-bit blocks
		0,0,2,4,6,12, // 2-5
		20,40,74,148,284, // 6-10
		568,1116,2232,4424,8848, // 11-15
		17622,35244,70340,140680,281076, // 16-20
		562152,1123736,2247472,4493828,8987656, // 21-25
		17973080,35946160,71887896,143775792,287542736,575085472 // 26-31
	};

	templ=new int[NOTMmax]; // m-bit template to test

	for(int m=NOTMmin; m<=NOTMmax; INCR(m,NOTMstp)) {

		int N=n/(100.*(1ul<<m)+m-1.); // N=8 in the original NIST's code (too weak)
		N=max(3,N); N=min(1000,N); const int M=n/N;

		const double Ew=(M-m+1)/double(1ul<<m), // Mean (counts in each bin)
			VARw=M*(1./(1ul<<m)-(2.*m-1.)/pow(2.,2.*m)); // Variance

		if(Ew<5) break;

		ULONG *const Wj=new ULONG[N];

		numOfTemplates=n_templ[m]; SKIP=1;
		if(NOTMtemp) { // if =0 all the templates will be tested
			if(numOfTemplates>NOTMtemp) { 
				SKIP=numOfTemplates/NOTMtemp; numOfTemplates=NOTMtemp; 
			}
		}
		skip=SKIP; V=0; for(int k=0;k<Bb;k++) A[k]=0; // Initialize CreateTemplate()

		for(int tc=0; tc<numOfTemplates; tc++) {

			if(m>2) CreateTemplate(m);
			else {
				if(tc) { templ[0]=1; templ[1]=0; } // 2nd template
				else   { templ[0]=0; templ[1]=1; } // 1st template
			}

			memset(Wj,0,N*sizeof(ULONG)); ULONG i;
			for(i = 0; i < N; i++) {
				for(ULONG j = 0; j <= M-m; j++) {
					int match = 1;
					for(int k = 0; k < m; k++)
						if(templ[k] != get_bit(i*M+j+k,bits)) {
							match = 0; break;
						}
					if (match) Wj[i]++;
				}
			}

			double XX=0; // Compute chi-square
			for(i=0;i<N;i++) { //printf("%d\t%d\n",i+1,Wj[i]);
				XX+=(Wj[i]-Ew)*(Wj[i]-Ew)/VARw;
			}

			int H=0; for(i=0;i<m;i++) H=H*2 + templ[i];
			sprintf(buf,"NOTM %d_%0X",m,H); SAVE(buf,chiQ(XX,N));
		}
		delete[] Wj;
	}
	delete[] templ;
}
