/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include "defi.h"

extern int
	PERMsmin,PERMsmax,PERMsstp, // Block size [bits]
	PERMtmin,PERMtmax,PERMtstp; // Group size [s-bit numbers]

// Permutation test [The Art of Computer Programming PAGE 64 (78/704)]

// WARNING: this test allocates n/block_size * 4 bytes

void PERM(void)
{
	ULONG *const u=new ULONG[PERMtmax+1]; // Temporary vector for U[b]..U[b+t]

	for(int s=PERMsmin; s<=PERMsmax; INCR(s,PERMsstp)) { // Block size [bits]

		const ULONG nn=n/s; // Number of numbers
		ULONG i,*const U=new ULONG[nn]; // s-bit numbers

		// Convert the s-bit blocks in nn numbers
		switch(s) { // ~76% faster (the U[i] are differents; bits in reverse order)
		case 32: for(i=0;i<nn;i++) U[i]=bits[i]; break;
		case 16: for(i=0;i<nn;i++) U[i]=((unsigned short*)bits)[i]; break;
		case  8: for(i=0;i<nn;i++) U[i]=((unsigned char*)bits)[i]; break;
		default:
			for(i=0;i<nn;i++) {
				ULONG u=0; for(int j=0;j<s;j++) u=u*2 + get_bit(i*s+j,bits);
				U[i]=u;
			}
		}

		// t: 11!=39,916,800  12!=479,001,600  13!=6,227,020,800
		for(int t=PERMtmin; t<=PERMtmax; INCR(t,PERMtstp)) { // Group size [s-bit numbers]

			ULONG fac=2; for(i=3;i<=t;i++) fac*=i; // Counter for the t! possible f's

			if(nn<5*fac*t) break; // At least 5 counts in each bin

			ULONG *const F=new ULONG[fac]; memset(F,0,fac*sizeof(ULONG));

			for(ULONG b=0;b<nn-t;b+=t) {

				int inv=0; // THE NUMBERS IN EACH t-number GROUP MUST BE UNIQUE

				for(i=0;i<t;i++) {
					for(ULONG j=b+i+1;j<b+t;j++)
						if(U[b+i]==U[j]) { inv=1; goto invalid; }

					u[i+1]=U[b+i]; // Copy the t numbers in the temporary vector
				}

				invalid: if(inv) continue; // Duplicate numbers

				ULONG r=t,f=0; // Order ID (0 <= f < t!)
				do { // Find the order ID
					ULONG Umax=0,Imax=0;
					for(i=1;i<=r;i++) if(u[i]>Umax) { Umax=u[i]; Imax=i; }
					f=r*f+Imax-1;
					ULONG t=u[r]; u[r]=u[Imax]; u[Imax]=t;
				} while(--r>1);

				F[f]++; // Count for the order ID
			}

			ULONG S=1; for(i=0;i<fac;i++) S+=F[i];
			double XX=0,E=S/double(fac);
			if(E>5) {
				for(i=0;i<fac;i++) XX+=(F[i]-E)*(F[i]-E)/E;
				sprintf(buf,"Permutation %d_%d",s,t); SAVE(buf,chiQ(XX,fac-1));
			}
			delete[] F;

		} // t

		delete[] U;

	} // s

	delete[] u;
}
