/* ---------------------------------------------------------------------------
 Copyright (c) 2004, Cristiano, Italy.   All rights reserved.

 LICENSE TERMS

 The free distribution and use of this software in both source and binary
 form is allowed (with or without changes) provided that:

   1. distributions of this source code include the above copyright
      notice, this list of conditions and the following disclaimer;

   2. distributions in binary form include the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other associated materials;

   3. the copyright holder's name is not used to endorse products
      built using this software without specific written permission.

 ALTERNATIVELY, provided that this notice is retained in full, this product
 may be distributed under the terms of the GNU General Public License (GPL),
 in which case the provisions of the GPL apply INSTEAD OF those given above.

 DISCLAIMER

 This software is provided 'as is' with no explicit or implied warranties
 in respect of its properties, including, but not limited to, correctness
 and/or fitness for purpose.

 --------------------------------------------------------------------------- */

#include <stdlib.h>
#include "defi.h"

/*
	printf("%08X\n",X);
	for(int i=0;i<32;i++) { printf("%d",get_bit(i,&X)); if((i+1)%4==0) printf(" "); }

	-- LITTLE ENDIAN --
X = 08DA7227
get_bit = 1110 0100 0100 1110 0101 1011 0001 0000
          7 <- 2 <- 2 <- 7 <- A <- D <- 8 <- 0    (read from right to left)

	-- BIG ENDIAN --
X = 2772DA08
get_bit = 0001 0000 0101 1011 0100 1110 1110 0100
          8    0    A    D    2    7    7    2
*/

int _FASTCALL_ get_bit(const ULONG i,const ULONG *x)
{ return (x[i>>5]>>(i&31)) & 1; }

void _FASTCALL_ set_bit(const ULONG i,const ULONG v,ULONG *const x)
{ v? x[i>>5] |= 1uL<<(i&31) : x[i>>5] &=~(1uL<<(i&31)); }

ULONG bpos;
ULONG _FASTCALL_ rng(void)
{
	//if(bpos>n>>5) printf("\n\t!! BPOS !!\n\n");
	return bits[bpos++];
}

double _FASTCALL_ chiQ(const double x,const int n)
{
	return igamc(double(n)/2.,x/2.);
}

/* error function in double precision */
double _FASTCALL_ erfc(double x)
{
	 double t, u, y;

	 t = 3.97886080735226 / (fabs(x) + 3.97886080735226);
	 u = t - 0.5;
	 y = (((((((((0.00127109764952614092 * u + 1.19314022838340944e-4) * u -
		  0.003963850973605135) * u - 8.70779635317295828e-4) * u +
		  0.00773672528313526668) * u + 0.00383335126264887303) * u -
		  0.0127223813782122755) * u - 0.0133823644533460069) * u +
		  0.0161315329733252248) * u + 0.0390976845588484035) * u +
		  0.00249367200053503304;
	 y = ((((((((((((y * u - 0.0838864557023001992) * u -
		  0.119463959964325415) * u + 0.0166207924969367356) * u +
		  0.357524274449531043) * u + 0.805276408752910567) * u +
		  1.18902982909273333) * u + 1.37040217682338167) * u +
		  1.31314653831023098) * u + 1.07925515155856677) * u +
		  0.774368199119538609) * u + 0.490165080585318424) * u +
		  0.275374741597376782) * t * exp(-x * x);
	 return x < 0 ? 2 - y : y;
}


// KS

#ifdef __GNUC__
int acmp(const void *a, const void *b)
#else
int __cdecl acmp(const void *a, const void *b)
#endif
{
	if (* (double *)a < *(double *)b) return -1;
	if (* (double *)a > *(double *)b) return 1;
	return 0;
}

void _FASTCALL_ sort_pv(double *x, const int n)
{
	qsort(x, n, sizeof(double),acmp);
}

static double _FASTCALL_ probks(double x)
{
	if(x < 0.2) return 1.0;
	int i = 0; double x2 = -2.0 * x * x, sum = 0.0, previous_sum = 0.0;

	while(++i) {
		sum = exp(x2 * i * i) - sum;
//		if(sum == previous_sum) break;
		if(fabs(sum-previous_sum)<1e-8) break;
		previous_sum = -sum;
	}
	return 2.0 * fabs(sum);
}


double _FASTCALL_ KStest(double *data,const int n) // Uniform distribution
{
	double en=n,ff,fn,fo=0,d=0; //sort_pv(data,n);
	for(int j=1;j<=n;j++) {
		fn=j/en; ff=data[j-1]; // CDF
		double f1=fabs(fo-ff),dt=fabs(fn-ff); if(f1>dt) dt=f1;
		fo=fn; if(dt>d) d=dt;
	}
	en=sqrt(en); return probks((en+0.12+0.11/en)*d);
}

double _FASTCALL_ KSnorm(double *data,const int n) // Normal distribution
{
	double en=n,ff,fn,fo=0,d=0; sort_pv(data,n);
	for(int j=1;j<=n;j++) {
		fn=j/en; ff=Phi(data[j-1]); // CDF
		double f1=fabs(fo-ff),dt=fabs(fn-ff); if(f1>dt) dt=f1;
		fo=fn; if(dt>d) d=dt;
	}
	en=sqrt(en); return probks((en+0.12+0.11/en)*d);
}

/*This test is based on a modified Kolmogorov-Smirnov method.
  The test-statistic is (FN(X)-X)**2/(X*(1-X)) (Anderson-Darling)
  where X is a uniform under null hypothesis. FN(X) is the empirical
  distribution of X.

!! EVEN WITH A PERFECT DISTRIBUTION AD<<1 (WHILE KStest=1) !!
*/

double _FASTCALL_ ADtest(double *y, const int n)
{
/*    TO TEST WHETHER A SET OF N REAL NUMBERS IS DRAWN
      FROM A UNIFORM DISTRIBUTION (KOLMOGOROV-SMIRNOV METHOD)
      THE TEST IS BASED ON THE DISTANCE BETWEEN THE EMPIRICAL
      AND THEORETICAL DISTRIBUTION FUNCTIONS
       USAGE: CALL KSTEST(Y,N,P)
      Y ...   ARRAY OF REAL NUMBERS HYPOTHETICALLY DRAWN
              FROM A UNIFORM DISTRIBUTION ON (0,1)
      N ...   NUMBER OF ELEMENTS IN 'Y'
      P IS THE PROBABILITY ASSOCIATED WITH THE OBSERVED VALUE
      OF THE ANDERSON-DARLING STATISTIC: N TIMES THE INTEGRAL
      OF (FN(X)-X)**2/(X*(1-X))
*/
	double z= - (double) n * n;
	for (int i = 0; i < n; ++i) {
		double t = y[i] * (1 - y[n-1 - i]); if (t < 1e-20) t = 1e-20;
		z -= (i + i + 1) * log(t);
	}
	z /= n; if(z<0.01) return 1;
	if(z<=2) return 1- exp(-1.2337/z)*2*(z/8+1-z*.04958*z/(z+1.325))/sqrt(z);
	if(z<=4) return exp(z*-1.091638)*.6621361+exp(z*-2.005138)*.95059;
	return exp(z*-1.050321)*.4938691+exp(z*-1.527198)*0.5946335;
}

double _FASTCALL_ SLtest(double *pv,const int Np)
{
	double em=0,et=0;
	for(int i=0;i<Np;i++) {
		double y=i/double(Np-1);
		em+=y*y; // Maximum error
		et+=(y-pv[i])*(y-pv[i]); // P-vals error
	}
	return et/em*100;
}
